from __future__ import annotations

from coredis._enum import CaseAndEncodingInsensitiveEnum


class PureToken(CaseAndEncodingInsensitiveEnum):
    """
    Enum for using pure-tokens with the redis api.
    """

    #: Used by:
    #:
    #:  - ``ACL LOG``
    RESET = b"RESET"

    #: Used by:
    #:
    #:  - ``BGSAVE``
    SCHEDULE = b"SCHEDULE"

    #: Used by:
    #:
    #:  - ``BITCOUNT``
    #:  - ``BITPOS``
    BIT = b"BIT"

    #: Used by:
    #:
    #:  - ``BITCOUNT``
    #:  - ``BITPOS``
    BYTE = b"BYTE"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    FAIL = b"FAIL"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    SAT = b"SAT"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    WRAP = b"WRAP"

    #: Used by:
    #:
    #:  - ``BITOP``
    AND = b"AND"

    #: Used by:
    #:
    #:  - ``BITOP``
    ANDOR = b"ANDOR"

    #: Used by:
    #:
    #:  - ``BITOP``
    DIFF = b"DIFF"

    #: Used by:
    #:
    #:  - ``BITOP``
    DIFF1 = b"DIFF1"

    #: Used by:
    #:
    #:  - ``BITOP``
    NOT = b"NOT"

    #: Used by:
    #:
    #:  - ``BITOP``
    #:  - ``REPLICAOF``
    #:  - ``SLAVEOF``
    ONE = b"ONE"

    #: Used by:
    #:
    #:  - ``BITOP``
    OR = b"OR"

    #: Used by:
    #:
    #:  - ``BITOP``
    XOR = b"XOR"

    #: Used by:
    #:
    #:  - ``BLMOVE``
    #:  - ``BLMPOP``
    #:  - ``LMOVE``
    #:  - ``LMPOP``
    LEFT = b"LEFT"

    #: Used by:
    #:
    #:  - ``BLMOVE``
    #:  - ``BLMPOP``
    #:  - ``LMOVE``
    #:  - ``LMPOP``
    RIGHT = b"RIGHT"

    #: Used by:
    #:
    #:  - ``BZMPOP``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZMPOP``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    MAX = b"MAX"

    #: Used by:
    #:
    #:  - ``BZMPOP``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZMPOP``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    MIN = b"MIN"

    #: Used by:
    #:
    #:  - ``CLIENT CACHING``
    #:  - ``CLIENT KILL``
    #:  - ``REPLICAOF``
    #:  - ``SCRIPT DEBUG``
    #:  - ``SLAVEOF``
    NO = b"NO"

    #: Used by:
    #:
    #:  - ``CLIENT CACHING``
    #:  - ``CLIENT KILL``
    #:  - ``SCRIPT DEBUG``
    YES = b"YES"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    MASTER = b"MASTER"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    NORMAL = b"NORMAL"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    PUBSUB = b"PUBSUB"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    REPLICA = b"REPLICA"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    SLAVE = b"SLAVE"

    #: Used by:
    #:
    #:  - ``CLIENT NO-EVICT``
    #:  - ``CLIENT NO-TOUCH``
    #:  - ``CLIENT REPLY``
    #:  - ``CLIENT TRACKING``
    OFF = b"OFF"

    #: Used by:
    #:
    #:  - ``CLIENT NO-EVICT``
    #:  - ``CLIENT NO-TOUCH``
    #:  - ``CLIENT REPLY``
    #:  - ``CLIENT TRACKING``
    ON = b"ON"

    #: Used by:
    #:
    #:  - ``CLIENT PAUSE``
    #:  - ``CLUSTER MIGRATION``
    ALL = b"ALL"

    #: Used by:
    #:
    #:  - ``CLIENT PAUSE``
    WRITE = b"WRITE"

    #: Used by:
    #:
    #:  - ``CLIENT REPLY``
    SKIP = b"SKIP"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    BCAST = b"BCAST"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    NOLOOP = b"NOLOOP"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    OPTIN = b"OPTIN"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    OPTOUT = b"OPTOUT"

    #: Used by:
    #:
    #:  - ``CLIENT UNBLOCK``
    ERROR = b"ERROR"

    #: Used by:
    #:
    #:  - ``CLIENT UNBLOCK``
    TIMEOUT = b"TIMEOUT"

    #: Used by:
    #:
    #:  - ``CLUSTER FAILOVER``
    #:  - ``FAILOVER``
    #:  - ``SHUTDOWN``
    #:  - ``XCLAIM``
    FORCE = b"FORCE"

    #: Used by:
    #:
    #:  - ``CLUSTER FAILOVER``
    TAKEOVER = b"TAKEOVER"

    #: Used by:
    #:
    #:  - ``CLUSTER RESET``
    HARD = b"HARD"

    #: Used by:
    #:
    #:  - ``CLUSTER RESET``
    SOFT = b"SOFT"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    STABLE = b"STABLE"

    #: Used by:
    #:
    #:  - ``CLUSTER SLOT-STATS``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    ASC = b"ASC"

    #: Used by:
    #:
    #:  - ``CLUSTER SLOT-STATS``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    DESC = b"DESC"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    SNAPSHOT_EOF = b"SNAPSHOT-EOF"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    STREAM_EOF = b"STREAM-EOF"

    #: Used by:
    #:
    #:  - ``COPY``
    #:  - ``FUNCTION LOAD``
    #:  - ``FUNCTION RESTORE``
    #:  - ``MIGRATE``
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    REPLACE = b"REPLACE"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``HEXPIRE``
    #:  - ``HEXPIREAT``
    #:  - ``HPEXPIRE``
    #:  - ``HPEXPIREAT``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``ZADD``
    GT = b"GT"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``HEXPIRE``
    #:  - ``HEXPIREAT``
    #:  - ``HPEXPIRE``
    #:  - ``HPEXPIREAT``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``ZADD``
    LT = b"LT"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``GEOADD``
    #:  - ``HEXPIRE``
    #:  - ``HEXPIREAT``
    #:  - ``HPEXPIRE``
    #:  - ``HPEXPIREAT``
    #:  - ``JSON.SET``
    #:  - ``MSETEX``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``SET``
    #:  - ``ZADD``
    NX = b"NX"

    #: Used by:
    #:
    #:  - ``EXPIRE``
    #:  - ``EXPIREAT``
    #:  - ``GEOADD``
    #:  - ``HEXPIRE``
    #:  - ``HEXPIREAT``
    #:  - ``HPEXPIRE``
    #:  - ``HPEXPIREAT``
    #:  - ``JSON.SET``
    #:  - ``MSETEX``
    #:  - ``PEXPIRE``
    #:  - ``PEXPIREAT``
    #:  - ``SET``
    #:  - ``ZADD``
    XX = b"XX"

    #: Used by:
    #:
    #:  - ``FAILOVER``
    #:  - ``SHUTDOWN``
    ABORT = b"ABORT"

    #: Used by:
    #:
    #:  - ``FLUSHALL``
    #:  - ``FLUSHDB``
    #:  - ``FUNCTION FLUSH``
    #:  - ``SCRIPT FLUSH``
    ASYNC = b"ASYNC"

    #: Used by:
    #:
    #:  - ``FLUSHALL``
    #:  - ``FLUSHDB``
    #:  - ``FUNCTION FLUSH``
    #:  - ``SCRIPT DEBUG``
    #:  - ``SCRIPT FLUSH``
    SYNC = b"SYNC"

    #: Used by:
    #:
    #:  - ``FUNCTION LIST``
    WITHCODE = b"WITHCODE"

    #: Used by:
    #:
    #:  - ``FUNCTION RESTORE``
    APPEND = b"APPEND"

    #: Used by:
    #:
    #:  - ``FUNCTION RESTORE``
    FLUSH = b"FLUSH"

    #: Used by:
    #:
    #:  - ``GEOADD``
    #:  - ``ZADD``
    CHANGE = b"CH"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    FT = b"FT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    KM = b"KM"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    M = b"M"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``GEODIST``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    MI = b"MI"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    ANY = b"ANY"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    WITHCOORD = b"WITHCOORD"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    WITHDIST = b"WITHDIST"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    WITHHASH = b"WITHHASH"

    #: Used by:
    #:
    #:  - ``GEOSEARCHSTORE``
    STOREDIST = b"STOREDIST"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``HGETEX``
    PERSIST = b"PERSIST"

    #: Used by:
    #:
    #:  - ``HOTKEYS START``
    CPU = b"CPU"

    #: Used by:
    #:
    #:  - ``HOTKEYS START``
    NET = b"NET"

    #: Used by:
    #:
    #:  - ``HRANDFIELD``
    WITHVALUES = b"WITHVALUES"

    #: Used by:
    #:
    #:  - ``HSCAN``
    NOVALUES = b"NOVALUES"

    #: Used by:
    #:
    #:  - ``HSETEX``
    FNX = b"FNX"

    #: Used by:
    #:
    #:  - ``HSETEX``
    FXX = b"FXX"

    #: Used by:
    #:
    #:  - ``HSETEX``
    #:  - ``MSETEX``
    #:  - ``SET``
    KEEPTTL = b"KEEPTTL"

    #: Used by:
    #:
    #:  - ``LCS``
    IDX = b"IDX"

    #: Used by:
    #:
    #:  - ``LCS``
    LEN = b"LEN"

    #: Used by:
    #:
    #:  - ``LCS``
    WITHMATCHLEN = b"WITHMATCHLEN"

    #: Used by:
    #:
    #:  - ``LINSERT``
    AFTER = b"AFTER"

    #: Used by:
    #:
    #:  - ``LINSERT``
    BEFORE = b"BEFORE"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    COPY = b"COPY"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    EMPTY_STRING = b""

    #: Used by:
    #:
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    ABSTTL = b"ABSTTL"

    #: Used by:
    #:
    #:  - ``SET``
    GET = b"GET"

    #: Used by:
    #:
    #:  - ``SHUTDOWN``
    NOSAVE = b"NOSAVE"

    #: Used by:
    #:
    #:  - ``SHUTDOWN``
    NOW = b"NOW"

    #: Used by:
    #:
    #:  - ``SHUTDOWN``
    SAVE = b"SAVE"

    #: Used by:
    #:
    #:  - ``SORT``
    #:  - ``SORT_RO``
    SORTING = b"ALPHA"

    #: Used by:
    #:
    #:  - ``XACKDEL``
    #:  - ``XADD``
    #:  - ``XDELEX``
    #:  - ``XTRIM``
    ACKED = b"ACKED"

    #: Used by:
    #:
    #:  - ``XACKDEL``
    #:  - ``XADD``
    #:  - ``XDELEX``
    #:  - ``XTRIM``
    DELREF = b"DELREF"

    #: Used by:
    #:
    #:  - ``XACKDEL``
    #:  - ``XADD``
    #:  - ``XDELEX``
    #:  - ``XTRIM``
    KEEPREF = b"KEEPREF"

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    APPROXIMATELY = b"~"

    #: Used by:
    #:
    #:  - ``XADD``
    AUTO_ID = b"*"

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    EQUAL = b"="

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    MAXLEN = b"MAXLEN"

    #: Used by:
    #:
    #:  - ``XADD``
    #:  - ``XTRIM``
    MINID = b"MINID"

    #: Used by:
    #:
    #:  - ``XADD``
    NOMKSTREAM = b"NOMKSTREAM"

    #: Used by:
    #:
    #:  - ``XAUTOCLAIM``
    #:  - ``XCLAIM``
    JUSTID = b"JUSTID"

    #: Used by:
    #:
    #:  - ``XGROUP CREATE``
    MKSTREAM = b"MKSTREAM"

    #: Used by:
    #:
    #:  - ``XGROUP CREATE``
    #:  - ``XGROUP SETID``
    NEW_ID = b"$"

    #: Used by:
    #:
    #:  - ``XINFO STREAM``
    FULL = b"FULL"

    #: Used by:
    #:
    #:  - ``XREADGROUP``
    NOACK = b"NOACK"

    #: Used by:
    #:
    #:  - ``ZADD``
    INCREMENT = b"INCR"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``FT.SUGGET``
    #:  - ``VLINKS``
    #:  - ``VSIM``
    #:  - ``ZDIFF``
    #:  - ``ZINTER``
    #:  - ``ZRANDMEMBER``
    #:  - ``ZRANGE``
    #:  - ``ZRANGEBYSCORE``
    #:  - ``ZREVRANGE``
    #:  - ``ZREVRANGEBYSCORE``
    #:  - ``ZUNION``
    WITHSCORES = b"WITHSCORES"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    SUM = b"SUM"

    #: Used by:
    #:
    #:  - ``ZRANGE``
    #:  - ``ZRANGESTORE``
    BYLEX = b"BYLEX"

    #: Used by:
    #:
    #:  - ``ZRANGE``
    #:  - ``ZRANGESTORE``
    BYSCORE = b"BYSCORE"

    #: Used by:
    #:
    #:  - ``ZRANGE``
    #:  - ``ZRANGESTORE``
    REV = b"REV"

    #: Used by:
    #:
    #:  - ``ZRANK``
    #:  - ``ZREVRANK``
    WITHSCORE = b"WITHSCORE"

    #: Used by:
    #:
    #:  - ``VADD``
    BIN = b"BIN"

    #: Used by:
    #:
    #:  - ``VADD``
    CAS = b"CAS"

    #: Used by:
    #:
    #:  - ``VADD``
    #:  - ``VSIM``
    FP32 = b"FP32"

    #: Used by:
    #:
    #:  - ``VADD``
    NOQUANT = b"NOQUANT"

    #: Used by:
    #:
    #:  - ``VADD``
    Q8 = b"Q8"

    #: Used by:
    #:
    #:  - ``VADD``
    #:  - ``VSIM``
    VALUES = b"VALUES"

    #: Used by:
    #:
    #:  - ``VSIM``
    ELE = b"ELE"

    #: Used by:
    #:
    #:  - ``VSIM``
    NOTHREAD = b"NOTHREAD"

    #: Used by:
    #:
    #:  - ``VSIM``
    TRUTH = b"TRUTH"

    #: Used by:
    #:
    #:  - ``VSIM``
    WITHATTRIBS = b"WITHATTRIBS"

    #: Used by:
    #:
    #:  - ``VEMB``
    RAW = b"RAW"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``BF.RESERVE``
    NONSCALING = b"NONSCALING"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    #:  - ``BF.INSERT``
    #:  - ``CF.INSERT``
    #:  - ``CF.INSERTNX``
    ITEMS = b"ITEMS"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``CF.INSERT``
    #:  - ``CF.INSERTNX``
    NOCREATE = b"NOCREATE"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    CAPACITY = b"CAPACITY"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    EXPANSION = b"EXPANSION"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    FILTERS = b"FILTERS"

    #: Used by:
    #:
    #:  - ``BF.INFO``
    SIZE = b"SIZE"

    #: Used by:
    #:
    #:  - ``CMS.MERGE``
    WEIGHTS = b"WEIGHTS"

    #: Used by:
    #:
    #:  - ``TOPK.LIST``
    WITHCOUNT = b"WITHCOUNT"

    #: Used by:
    #:
    #:  - ``TDIGEST.MERGE``
    COMPRESSION = b"COMPRESSION"

    #: Used by:
    #:
    #:  - ``TDIGEST.MERGE``
    OVERRIDE = b"OVERRIDE"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    BLOCK = b"BLOCK"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.CREATE``
    COMPRESSED = b"COMPRESSED"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    FIRST = b"FIRST"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    LAST = b"LAST"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    UNCOMPRESSED = b"UNCOMPRESSED"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    AVG = b"AVG"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    COUNT = b"COUNT"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    COUNTALL = b"COUNTALL"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    COUNTNAN = b"COUNTNAN"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    RANGE = b"RANGE"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    STD_P = b"STD.P"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    STD_S = b"STD.S"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    TWA = b"TWA"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    VAR_P = b"VAR.P"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    VAR_S = b"VAR.S"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    BUCKETTIMESTAMP = b"BUCKETTIMESTAMP"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    EMPTY = b"EMPTY"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    FILTER_BY_VALUE = b"FILTER_BY_VALUE"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    END = b"END"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    HYPHEN_MINUS = b"-"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    MID = b"MID"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    PLUS_SIGN = b"+"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    START = b"START"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    TILDE = b"~"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    GROUPBY = b"GROUPBY"

    #: Used by:
    #:
    #:  - ``TS.MGET``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    SELECTED_LABELS = b"SELECTED_LABELS"

    #: Used by:
    #:
    #:  - ``TS.MGET``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    WITHLABELS = b"WITHLABELS"

    #: Used by:
    #:
    #:  - ``TS.MREVRANGE``
    LATEST = b"LATEST"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    DISABLE = b"DISABLE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    ENABLE = b"ENABLE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    GEO = b"GEO"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    GEOSHAPE = b"GEOSHAPE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    HASH = b"HASH"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    INDEXEMPTY = b"INDEXEMPTY"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    INDEXMISSING = b"INDEXMISSING"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    JSON = b"JSON"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    MAXTEXTFIELDS = b"MAXTEXTFIELDS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOFIELDS = b"NOFIELDS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOFREQS = b"NOFREQS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOHL = b"NOHL"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOINDEX = b"NOINDEX"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NOOFFSETS = b"NOOFFSETS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    NUMERIC = b"NUMERIC"

    #: Used by:
    #:
    #:  - ``FT.ALTER``
    #:  - ``FT.CREATE``
    SCHEMA = b"SCHEMA"

    #: Used by:
    #:
    #:  - ``FT.ALTER``
    #:  - ``FT.CREATE``
    #:  - ``FT.SYNUPDATE``
    SKIPINITIALSCAN = b"SKIPINITIALSCAN"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    SORTABLE = b"SORTABLE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    TAG = b"TAG"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    TEXT = b"TEXT"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    UNF = b"UNF"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    VECTOR = b"VECTOR"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    WITHSUFFIXTRIE = b"WITHSUFFIXTRIE"

    #: Used by:
    #:
    #:  - ``FT.ALTER``
    ADD = b"ADD"

    #: Used by:
    #:
    #:  - ``FT.DROPINDEX``
    DELETE_DOCS = b"DD"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    EXCLUDE = b"EXCLUDE"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    INCLUDE = b"INCLUDE"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    EXPLAINSCORE = b"EXPLAINSCORE"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    HIGHLIGHT = b"HIGHLIGHT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    INORDER = b"INORDER"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    LIMIT = b"LIMIT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    NOCONTENT = b"NOCONTENT"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    NOSTOPWORDS = b"NOSTOPWORDS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    PARAMS = b"PARAMS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SUMMARIZE = b"SUMMARIZE"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    TAGS = b"TAGS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SEARCH``
    VERBATIM = b"VERBATIM"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``FT.SUGGET``
    WITHPAYLOADS = b"WITHPAYLOADS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    WITHSORTKEYS = b"WITHSORTKEYS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    COUNT_DISTINCT = b"COUNT_DISTINCT"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    COUNT_DISTINCTISH = b"COUNT_DISTINCTISH"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    FIRST_VALUE = b"FIRST_VALUE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    LOADALL = b"LOAD *"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    QUANTILE = b"QUANTILE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    RANDOM_SAMPLE = b"RANDOM_SAMPLE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    REDUCE = b"REDUCE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    STDDEV = b"STDDEV"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    TOLIST = b"TOLIST"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    WITHCURSOR = b"WITHCURSOR"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    AGGREGATE = b"AGGREGATE"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    HYBRID = b"HYBRID"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    LIMITED = b"LIMITED"

    #: Used by:
    #:
    #:  - ``FT.PROFILE``
    QUERYWORD = b"QUERY"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    #:  - ``FT.PROFILE``
    SEARCH = b"SEARCH"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    ADHOC = b"ADHOC"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    BATCHES = b"BATCHES"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    COMBINE = b"COMBINE"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    KNN = b"KNN"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    LINEAR = b"LINEAR"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    NOSORT = b"NOSORT"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    RRF = b"RRF"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    VSIM = b"VSIM"

    #: Used by:
    #:
    #:  - ``FT.SUGADD``
    INCR = b"INCR"

    #: Used by:
    #:
    #:  - ``FT.SUGGET``
    FUZZY = b"FUZZY"


class PrefixToken(CaseAndEncodingInsensitiveEnum):
    """
    Enum for internal use when adding prefixes to arguments
    """

    #: Used by:
    #:
    #:  - ``BITFIELD``
    #:  - ``BITFIELD_RO``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    GET = b"GET"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    INCRBY = b"INCRBY"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    OVERFLOW = b"OVERFLOW"

    #: Used by:
    #:
    #:  - ``BITFIELD``
    SET = b"SET"

    #: Used by:
    #:
    #:  - ``BLMPOP``
    #:  - ``BZMPOP``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.CURSOR READ``
    #:  - ``FT.HYBRID``
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``GEORADIUSBYMEMBER_RO``
    #:  - ``GEORADIUS_RO``
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    #:  - ``HOTKEYS START``
    #:  - ``HSCAN``
    #:  - ``LMPOP``
    #:  - ``LPOS``
    #:  - ``SCAN``
    #:  - ``SSCAN``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    #:  - ``VSIM``
    #:  - ``XAUTOCLAIM``
    #:  - ``XINFO STREAM``
    #:  - ``XRANGE``
    #:  - ``XREAD``
    #:  - ``XREADGROUP``
    #:  - ``XREVRANGE``
    #:  - ``ZMPOP``
    #:  - ``ZSCAN``
    COUNT = b"COUNT"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    ADDR = b"ADDR"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    #:  - ``CLUSTER MIGRATION``
    IDENTIFIER = b"ID"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    LADDR = b"LADDR"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    MAXAGE = b"MAXAGE"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    SKIPME = b"SKIPME"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    #:  - ``CLIENT LIST``
    #:  - ``SCAN``
    TYPE = b"TYPE"

    #: Used by:
    #:
    #:  - ``CLIENT KILL``
    USER = b"USER"

    #: Used by:
    #:
    #:  - ``CLIENT SETINFO``
    LIB_NAME = b"LIB-NAME"

    #: Used by:
    #:
    #:  - ``CLIENT SETINFO``
    LIB_VER = b"LIB-VER"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    #:  - ``FT.CREATE``
    PREFIX = b"PREFIX"

    #: Used by:
    #:
    #:  - ``CLIENT TRACKING``
    REDIRECT = b"REDIRECT"

    #: Used by:
    #:
    #:  - ``CLUSTER MIGRATION``
    CANCEL = b"CANCEL"

    #: Used by:
    #:
    #:  - ``CLUSTER MIGRATION``
    IMPORT = b"IMPORT"

    #: Used by:
    #:
    #:  - ``CLUSTER MIGRATION``
    STATUS = b"STATUS"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    IMPORTING = b"IMPORTING"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    MIGRATING = b"MIGRATING"

    #: Used by:
    #:
    #:  - ``CLUSTER SETSLOT``
    NODE = b"NODE"

    #: Used by:
    #:
    #:  - ``CLUSTER SLOT-STATS``
    #:  - ``SINTERCARD``
    #:  - ``SORT``
    #:  - ``SORT_RO``
    #:  - ``XADD``
    #:  - ``XTRIM``
    #:  - ``ZINTERCARD``
    #:  - ``ZRANGE``
    #:  - ``ZRANGEBYLEX``
    #:  - ``ZRANGEBYSCORE``
    #:  - ``ZRANGESTORE``
    #:  - ``ZREVRANGEBYLEX``
    #:  - ``ZREVRANGEBYSCORE``
    LIMIT = b"LIMIT"

    #: Used by:
    #:
    #:  - ``CLUSTER SLOT-STATS``
    ORDERBY = b"ORDERBY"

    #: Used by:
    #:
    #:  - ``CLUSTER SLOT-STATS``
    SLOTSRANGE = b"SLOTSRANGE"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    ACK = b"ACK"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    CONF = b"CONF"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    FAIL = b"FAIL"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    RDBCHANNEL = b"RDBCHANNEL"

    #: Used by:
    #:
    #:  - ``CLUSTER SYNCSLOTS``
    SYNC = b"SYNC"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    ACLCAT = b"ACLCAT"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    FILTERBY = b"FILTERBY"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    MODULE = b"MODULE"

    #: Used by:
    #:
    #:  - ``COMMAND LIST``
    PATTERN = b"PATTERN"

    #: Used by:
    #:
    #:  - ``COPY``
    DB = b"DB"

    #: Used by:
    #:
    #:  - ``DELEX``
    #:  - ``SET``
    IFDEQ = b"IFDEQ"

    #: Used by:
    #:
    #:  - ``DELEX``
    #:  - ``SET``
    IFDNE = b"IFDNE"

    #: Used by:
    #:
    #:  - ``DELEX``
    #:  - ``SET``
    IFEQ = b"IFEQ"

    #: Used by:
    #:
    #:  - ``DELEX``
    #:  - ``SET``
    IFNE = b"IFNE"

    #: Used by:
    #:
    #:  - ``FAILOVER``
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    TIMEOUT = b"TIMEOUT"

    #: Used by:
    #:
    #:  - ``FAILOVER``
    TO = b"TO"

    #: Used by:
    #:
    #:  - ``FUNCTION LIST``
    LIBRARYNAME = b"LIBRARYNAME"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    #:  - ``SORT``
    STORE = b"STORE"

    #: Used by:
    #:
    #:  - ``GEORADIUS``
    #:  - ``GEORADIUSBYMEMBER``
    STOREDIST = b"STOREDIST"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    BYBOX = b"BYBOX"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    BYRADIUS = b"BYRADIUS"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    FROMLONLAT = b"FROMLONLAT"

    #: Used by:
    #:
    #:  - ``GEOSEARCH``
    #:  - ``GEOSEARCHSTORE``
    FROMMEMBER = b"FROMMEMBER"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``HGETEX``
    #:  - ``HSETEX``
    #:  - ``MSETEX``
    #:  - ``SET``
    EX = b"EX"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``HGETEX``
    #:  - ``HSETEX``
    #:  - ``MSETEX``
    #:  - ``SET``
    EXAT = b"EXAT"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``HGETEX``
    #:  - ``HSETEX``
    #:  - ``MSETEX``
    #:  - ``SET``
    PX = b"PX"

    #: Used by:
    #:
    #:  - ``GETEX``
    #:  - ``HGETEX``
    #:  - ``HSETEX``
    #:  - ``MSETEX``
    #:  - ``SET``
    PXAT = b"PXAT"

    #: Used by:
    #:
    #:  - ``HELLO``
    #:  - ``MIGRATE``
    AUTH = b"AUTH"

    #: Used by:
    #:
    #:  - ``HELLO``
    SETNAME = b"SETNAME"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``HEXPIRE``
    #:  - ``HEXPIREAT``
    #:  - ``HEXPIRETIME``
    #:  - ``HGETDEL``
    #:  - ``HGETEX``
    #:  - ``HPERSIST``
    #:  - ``HPEXPIRE``
    #:  - ``HPEXPIREAT``
    #:  - ``HPEXPIRETIME``
    #:  - ``HPTTL``
    #:  - ``HSETEX``
    #:  - ``HTTL``
    FIELDS = b"FIELDS"

    #: Used by:
    #:
    #:  - ``HOTKEYS START``
    DURATION = b"DURATION"

    #: Used by:
    #:
    #:  - ``HOTKEYS START``
    METRICS = b"METRICS"

    #: Used by:
    #:
    #:  - ``HOTKEYS START``
    SAMPLE = b"SAMPLE"

    #: Used by:
    #:
    #:  - ``HOTKEYS START``
    SLOTS = b"SLOTS"

    #: Used by:
    #:
    #:  - ``HSCAN``
    #:  - ``SCAN``
    #:  - ``SSCAN``
    #:  - ``ZSCAN``
    MATCH = b"MATCH"

    #: Used by:
    #:
    #:  - ``LCS``
    MINMATCHLEN = b"MINMATCHLEN"

    #: Used by:
    #:
    #:  - ``LOLWUT``
    VERSION = b"VERSION"

    #: Used by:
    #:
    #:  - ``LPOS``
    MAXLEN = b"MAXLEN"

    #: Used by:
    #:
    #:  - ``LPOS``
    RANK = b"RANK"

    #: Used by:
    #:
    #:  - ``MEMORY USAGE``
    SAMPLES = b"SAMPLES"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    AUTH2 = b"AUTH2"

    #: Used by:
    #:
    #:  - ``MIGRATE``
    KEYS = b"KEYS"

    #: Used by:
    #:
    #:  - ``MODULE LOADEX``
    ARGS = b"ARGS"

    #: Used by:
    #:
    #:  - ``MODULE LOADEX``
    CONFIG = b"CONFIG"

    #: Used by:
    #:
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    FREQ = b"FREQ"

    #: Used by:
    #:
    #:  - ``RESTORE``
    #:  - ``RESTORE-ASKING``
    IDLETIME = b"IDLETIME"

    #: Used by:
    #:
    #:  - ``SORT``
    #:  - ``SORT_RO``
    BY = b"BY"

    #: Used by:
    #:
    #:  - ``TRIMSLOTS``
    RANGES = b"RANGES"

    #: Used by:
    #:
    #:  - ``XACKDEL``
    #:  - ``XDELEX``
    IDS = b"IDS"

    #: Used by:
    #:
    #:  - ``XADD``
    IDMP = b"IDMP"

    #: Used by:
    #:
    #:  - ``XADD``
    IDMPAUTO = b"IDMPAUTO"

    #: Used by:
    #:
    #:  - ``XCFGSET``
    IDMP_DURATION = b"IDMP-DURATION"

    #: Used by:
    #:
    #:  - ``XCFGSET``
    IDMP_MAXSIZE = b"IDMP-MAXSIZE"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    #:  - ``XPENDING``
    IDLE = b"IDLE"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    LASTID = b"LASTID"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    RETRYCOUNT = b"RETRYCOUNT"

    #: Used by:
    #:
    #:  - ``XCLAIM``
    TIME = b"TIME"

    #: Used by:
    #:
    #:  - ``XGROUP CREATE``
    #:  - ``XGROUP SETID``
    ENTRIESREAD = b"ENTRIESREAD"

    #: Used by:
    #:
    #:  - ``XREAD``
    #:  - ``XREADGROUP``
    BLOCK = b"BLOCK"

    #: Used by:
    #:
    #:  - ``XREAD``
    #:  - ``XREADGROUP``
    STREAMS = b"STREAMS"

    #: Used by:
    #:
    #:  - ``XREADGROUP``
    CLAIM = b"CLAIM"

    #: Used by:
    #:
    #:  - ``XREADGROUP``
    GROUP = b"GROUP"

    #: Used by:
    #:
    #:  - ``XSETID``
    ENTRIESADDED = b"ENTRIESADDED"

    #: Used by:
    #:
    #:  - ``XSETID``
    MAXDELETEDID = b"MAXDELETEDID"

    #: Used by:
    #:
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    AGGREGATE = b"AGGREGATE"

    #: Used by:
    #:
    #:  - ``ZINTER``
    #:  - ``ZINTERSTORE``
    #:  - ``ZUNION``
    #:  - ``ZUNIONSTORE``
    WEIGHTS = b"WEIGHTS"

    #: Used by:
    #:
    #:  - ``VADD``
    #:  - ``VSIM``
    EF = b"EF"

    #: Used by:
    #:
    #:  - ``VADD``
    M = b"M"

    #: Used by:
    #:
    #:  - ``VADD``
    REDUCE = b"REDUCE"

    #: Used by:
    #:
    #:  - ``VADD``
    SETATTR = b"SETATTR"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    #:  - ``VSIM``
    EPSILON = b"EPSILON"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.CREATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    #:  - ``TS.MGET``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``VSIM``
    FILTER = b"FILTER"

    #: Used by:
    #:
    #:  - ``VSIM``
    FILTER_EF = b"FILTER-EF"

    #: Used by:
    #:
    #:  - ``JSON.GET``
    INDENT = b"INDENT"

    #: Used by:
    #:
    #:  - ``JSON.GET``
    NEWLINE = b"NEWLINE"

    #: Used by:
    #:
    #:  - ``JSON.GET``
    SPACE = b"SPACE"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``BF.RESERVE``
    #:  - ``CF.RESERVE``
    EXPANSION = b"EXPANSION"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    #:  - ``CF.INSERT``
    #:  - ``CF.INSERTNX``
    CAPACITY = b"CAPACITY"

    #: Used by:
    #:
    #:  - ``BF.INSERT``
    ERROR = b"ERROR"

    #: Used by:
    #:
    #:  - ``CF.RESERVE``
    BUCKETSIZE = b"BUCKETSIZE"

    #: Used by:
    #:
    #:  - ``CF.RESERVE``
    MAXITERATIONS = b"MAXITERATIONS"

    #: Used by:
    #:
    #:  - ``TDIGEST.CREATE``
    COMPRESSION = b"COMPRESSION"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    CHUNK_SIZE = b"CHUNK_SIZE"

    #: Used by:
    #:
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    DUPLICATE_POLICY = b"DUPLICATE_POLICY"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.CREATE``
    ENCODING = b"ENCODING"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    LABELS = b"LABELS"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    #:  - ``TS.ALTER``
    #:  - ``TS.CREATE``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    RETENTION = b"RETENTION"

    #: Used by:
    #:
    #:  - ``TS.ADD``
    ON_DUPLICATE = b"ON_DUPLICATE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``TS.DECRBY``
    #:  - ``TS.INCRBY``
    TIMESTAMP = b"TIMESTAMP"

    #: Used by:
    #:
    #:  - ``TS.CREATERULE``
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    AGGREGATION = b"AGGREGATION"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    ALIGN = b"ALIGN"

    #: Used by:
    #:
    #:  - ``TS.MRANGE``
    #:  - ``TS.MREVRANGE``
    #:  - ``TS.RANGE``
    #:  - ``TS.REVRANGE``
    FILTER_BY_TS = b"FILTER_BY_TS"

    #: Used by:
    #:
    #:  - ``TS.REVRANGE``
    BUCKETTIMESTAMP = b"BUCKETTIMESTAMP"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.CREATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    AS = b"AS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    INDEXALL = b"INDEXALL"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    #:  - ``FT.SEARCH``
    LANGUAGE = b"LANGUAGE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    LANGUAGE_FIELD = b"LANGUAGE_FIELD"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    ON = b"ON"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    PAYLOAD_FIELD = b"PAYLOAD_FIELD"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    SCORE = b"SCORE"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    SCORE_FIELD = b"SCORE_FIELD"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    STOPWORDS = b"STOPWORDS"

    #: Used by:
    #:
    #:  - ``FT.CREATE``
    TEMPORARY = b"TEMPORARY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.EXPLAIN``
    #:  - ``FT.EXPLAINCLI``
    #:  - ``FT.SEARCH``
    #:  - ``FT.SPELLCHECK``
    DIALECT = b"DIALECT"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    DISTANCE = b"DISTANCE"

    #: Used by:
    #:
    #:  - ``FT.SPELLCHECK``
    TERMS = b"TERMS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    EXPANDER = b"EXPANDER"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    FRAGS = b"FRAGS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    GEOFILTER = b"GEOFILTER"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    INFIELDS = b"INFIELDS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    INKEYS = b"INKEYS"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    LEN = b"LEN"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    #:  - ``FT.SUGADD``
    PAYLOAD = b"PAYLOAD"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    RETURN = b"RETURN"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    SCORER = b"SCORER"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SEPARATOR = b"SEPARATOR"

    #: Used by:
    #:
    #:  - ``FT.SEARCH``
    SLOP = b"SLOP"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    #:  - ``FT.SEARCH``
    SORTBY = b"SORTBY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    ABS = b"ABS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    APPLY = b"APPLY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    CEIL = b"CEIL"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    CONTAINS = b"CONTAINS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    DAY = b"DAY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    DAYOFMONTH = b"DAYOFMONTH"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    DAYOFWEEK = b"DAYOFWEEK"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    DAYOFYEAR = b"DAYOFYEAR"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    EXISTS = b"EXISTS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    EXP = b"EXP"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    FLOOR = b"FLOOR"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    FMT = b"FMT"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    FORMAT = b"FORMAT"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    GEODISTANCE = b"GEODISTANCE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    GROUPBY = b"GROUPBY"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    HOUR = b"HOUR"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    LOAD = b"LOAD"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    LOG = b"LOG"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    LOG2 = b"LOG2"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    LOWER = b"LOWER"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    MATCHED_TERMS = b"MATCHED_TERMS"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.SUGGET``
    MAX = b"MAX"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    MAX_TERMS_100 = b"MAX_TERMS=100"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    MAXIDLE = b"MAXIDLE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    MINUTE = b"MINUTE"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    MONTH = b"MONTH"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    MONTHOFYEAR = b"MONTHOFYEAR"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    OFFSET = b"OFFSET"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    PARSETIME = b"PARSETIME"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    S = b"S"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    S1 = b"S1"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    S2 = b"S2"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    SPLIT = b"SPLIT"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    SQRT = b"SQRT"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    STARTSWITH = b"STARTSWITH"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    STRLEN = b"STRLEN"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    SUBSTR = b"SUBSTR"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    TIMEFMT = b"TIMEFMT"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    TIMESHARING = b"TIMESHARING"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    UPPER = b"UPPER"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    X = b"X"

    #: Used by:
    #:
    #:  - ``FT.AGGREGATE``
    #:  - ``FT.HYBRID``
    YEAR = b"YEAR"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    ALPHA = b"ALPHA"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    BATCH_SIZE = b"BATCH_SIZE"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    BETA = b"BETA"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    CONSTANT = b"CONSTANT"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    EF_RUNTIME = b"EF_RUNTIME"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    K = b"K"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    POLICY = b"POLICY"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    RADIUS = b"RADIUS"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    SHARD_K_RATIO = b"SHARD_K_RATIO"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    WINDOW = b"WINDOW"

    #: Used by:
    #:
    #:  - ``FT.HYBRID``
    YIELD_SCORE_AS = b"YIELD_SCORE_AS"



from __future__ import annotations

import warnings

from ..patterns.pubsub import ClusterPubSub, PubSub, ShardedPubSub, SubscriptionCallback

warnings.warn(
    "coredis.commands.pubsub has been moved to coredis.patterns.pubsub",
    DeprecationWarning,
    stacklevel=2,
)
__all__ = ["PubSub", "ClusterPubSub", "ShardedPubSub", "SubscriptionCallback"]



from __future__ import annotations

import dataclasses
import functools
import inspect
import math
import os
import ssl
from abc import ABC, abstractmethod
from collections import deque
from typing import cast

from anyio import (
    TASK_STATUS_IGNORED,
    BrokenResourceError,
    CancelScope,
    CapacityLimiter,
    ClosedResourceError,
    EndOfStream,
    create_memory_object_stream,
    create_task_group,
    fail_after,
)
from anyio.abc import ByteStream, TaskGroup, TaskStatus
from anyio.lowlevel import checkpoint
from exceptiongroup import BaseExceptionGroup, catch

import coredis
from coredis._packer import Packer
from coredis._utils import logger, nativestr
from coredis.commands.constants import CommandName
from coredis.constants.resp import DataType
from coredis.credentials import AbstractCredentialProvider, UserPassCredentialProvider
from coredis.exceptions import ConnectionError, UnknownCommandError
from coredis.parser import NotEnoughData, Parser
from coredis.tokens import PureToken
from coredis.typing import (
    AsyncGenerator,
    Awaitable,
    Callable,
    Concatenate,
    NotRequired,
    P,
    R,
    RedisError,
    RedisValueT,
    ResponseType,
    Self,
    TypedDict,
    TypeVar,
)

from ._request import Request
from ._statistics import ConnectionStatistics

CERT_REQS = {
    "none": ssl.CERT_NONE,
    "optional": ssl.CERT_OPTIONAL,
    "required": ssl.CERT_REQUIRED,
}

ConnectionT = TypeVar("ConnectionT", bound="BaseConnection")


@dataclasses.dataclass(unsafe_hash=True)
class Location(ABC):
    """
    Abstract location
    """

    ...

    @abstractmethod
    async def check(self) -> bool:
        """
        Returns whether the location can be connected to
        """
        ...


class BaseConnectionParams(TypedDict):
    """
    The common parameters accepted by :class:`coredis.connection.BaseConnection`
    """

    #: Maximum time to wait for receiving a response
    #: for requests created through this connection.
    stream_timeout: NotRequired[float | None]
    #: Maximum time to wait for establishing a connection
    connect_timeout: NotRequired[float | None]

    #: Default encoding for command responses.
    encoding: NotRequired[str]
    #: Whether to automatically decode responses.
    decode_responses: NotRequired[bool]

    #: Optional name to register with the server.
    client_name: NotRequired[str | None]
    #: If True, disables replies for all commands.
    noreply: NotRequired[bool]
    #: If True, sets CLIENT NO-EVICT on the connection.
    noevict: NotRequired[bool]
    #: If True, sets CLIENT NO-TOUCH on the connection.
    notouch: NotRequired[bool]
    #: Maximum idle time in seconds before the connection is closed.
    max_idle_time: NotRequired[int | None]

    #: Limiter to throttle CPU-bound processing.
    processing_budget: NotRequired[CapacityLimiter]

    #: The username to use for authenticating against the redis server
    username: NotRequired[str | None]
    #: The password to use for authenticating against the redis server
    password: NotRequired[str | None]
    #: If provided the connection handshake will include authentication using this provider.
    credential_provider: NotRequired[AbstractCredentialProvider | None]
    #: If provided the connection will immediately switch to this db as part of the handshake
    db: NotRequired[int | None]
    #: For TLS connections, the ssl context to use when performing the TLS handshake
    ssl_context: NotRequired[ssl.SSLContext | None]


class RedisSSLContext:
    context: ssl.SSLContext | None

    def __init__(
        self,
        keyfile: str | None,
        certfile: str | None,
        cert_reqs: str | ssl.VerifyMode | None = None,
        ca_certs: str | None = None,
        check_hostname: bool | None = None,
    ) -> None:
        self.keyfile = keyfile
        self.certfile = certfile
        self.check_hostname = check_hostname if check_hostname is not None else False
        if cert_reqs is None:
            self.cert_reqs = ssl.CERT_OPTIONAL
        elif isinstance(cert_reqs, str):
            self.cert_reqs = CERT_REQS[cert_reqs]
        else:
            self.cert_reqs = cert_reqs
        self.ca_certs = ca_certs
        self.context = None

    def get(self) -> ssl.SSLContext:
        if not self.context:
            self.context = ssl.create_default_context()
            if self.certfile and self.keyfile:
                self.context.load_cert_chain(certfile=self.certfile, keyfile=self.keyfile)
            if self.ca_certs:
                self.context.load_verify_locations(
                    **{("capath" if os.path.isdir(self.ca_certs) else "cafile"): self.ca_certs}
                )
            self.context.check_hostname = self.check_hostname
            self.context.verify_mode = self.cert_reqs
        return self.context


@dataclasses.dataclass
class CommandInvocation:
    command: bytes
    args: tuple[RedisValueT, ...]
    decode: bool | None
    encoding: str | None


class BaseConnection(ABC):
    """
    Base class for Redis connections.

    Manages a low-level connection to a single Redis server:
    sending commands, queuing requests, receiving and parsing responses,
    handling RESP3 push messages, and connection lifecycle management.

    Subclasses must implement the :meth:`_connect` method to establish the underlying
    transport (TCP, UNIX socket, etc.).

    """

    Params = BaseConnectionParams
    """
    :meta private:
    """

    @staticmethod
    def _ensure_usable(
        function: Callable[Concatenate[ConnectionT, P], R],
    ) -> Callable[Concatenate[ConnectionT, P], R]:
        @functools.wraps(function)
        def _connection_ensured(slf: ConnectionT, /, *args: P.args, **kwargs: P.kwargs) -> R:
            if (not slf._handshake_attempted and slf.transport_healthy) or slf.usable:
                return function(slf, *args, **kwargs)
            raise ConnectionError("Connection not usable") from slf._last_error

        _connection_ensured.__doc__ = f"""{_connection_ensured.__doc__}
:raises: :exc:`coredis.exceptions.ConnectionError` if the connection is
 not usable.
"""
        return _connection_ensured

    def __init__(
        self,
        location: Location,
        *,
        stream_timeout: float | None = None,
        connect_timeout: float | None = None,
        max_idle_time: int | None = None,
        encoding: str = "utf-8",
        decode_responses: bool = False,
        credential_provider: AbstractCredentialProvider | None = None,
        username: str | None = None,
        password: str | None = None,
        client_name: str | None = None,
        db: int | None = 0,
        noreply: bool = False,
        noevict: bool = False,
        notouch: bool = False,
        ssl_context: ssl.SSLContext | None = None,
        processing_budget: CapacityLimiter | None = None,
    ):
        """
        :param location: The location of the server this connection is connecting to
        :param stream_timeout: Maximum time to wait for receiving a response
         for requests created through this connection.
        :param connect_timeout: Maximum time to wait for establishing a connection
        :param max_idle_time: Maximum idle time in seconds before the connection is closed.
        :param encoding: Default encoding for command responses.
        :param decode_responses: Whether to automatically decode responses.
        :param credential_provider: If provided the connection handshake will include
         authentication using this provider.
        :param username: The username to use for authenticating against the redis server
        :param password: The password to use for authenticating against the redis server
        :param client_name: Optional name to register with the server.
        :param db: If provided the connection will immediately switch to this db as part
         of the handshake
        :param noreply: If True, disables replies for all commands.
        :param noevict: If True, sets CLIENT NO-EVICT on the connection.
        :param notouch: If True, sets CLIENT NO-TOUCH on the connection.
        :param ssl_context: For TLS connections, the ssl context to use when performing
         the TLS handshake.
        :param processing_budget: limiter to throttle CPU-bound processing.
        """
        self.location = location
        self.statistics = ConnectionStatistics()

        self._stream_timeout = stream_timeout
        self._connect_timeout = connect_timeout
        # maximum time to wait for data from the server before
        # closing the connection
        self._max_idle_time = max_idle_time

        self._username = username
        self._password = password
        self._credential_provider = credential_provider

        self._db = db

        self._encoding = encoding
        self._decode_responses = decode_responses

        self._connect_callbacks: list[
            (Callable[[Self], Awaitable[None]] | Callable[[Self], None])
        ] = list()

        # server version as reported by the server
        self.server_version: str | None = None
        # name used to identify thie connection with the redis server
        self.client_name = client_name
        # id for this connection as returned by the redis server
        self.client_id: int | None = None
        # client id that the redis server should send any redirected notifications to
        self.tracking_client_id: int | None = None

        self._noreply = noreply
        self._noreply_set = False
        self._noevict = noevict
        self._notouch = notouch

        self._ssl_context = ssl_context

        self._task_group: TaskGroup | None = None
        # The actual connection to the server
        self.stream: ByteStream | None = None
        # buffer for push message types
        self._push_message_buffer_in, self._push_message_buffer_out = create_memory_object_stream[
            list[ResponseType]
        ](math.inf)
        # buffer for streaming messages
        self._streamed_message_buffer_in, self._streamed_message_buffer_out = (
            create_memory_object_stream[ResponseType](math.inf)
        )
        # RESP parser/packer
        #  for writes to the socket
        self._write_buffer_in, self._write_buffer_out = create_memory_object_stream[list[Request]](
            math.inf
        )
        self._parser = Parser()
        self._packer: Packer = Packer(self._encoding)

        self._requests: deque[Request] = deque()
        self._connection_cancel_scope: CancelScope | None = None

        # Error & State flags
        self._last_error: BaseException | None = None
        self._ready = False
        self._handshake_attempted = False
        self._transport_failed = False
        self._terminated = False

        # To be used in the read task for cpu bound processing after data is received
        self._processing_budget = processing_budget or CapacityLimiter(1)

    def __repr__(self) -> str:
        return self.describe()

    @abstractmethod
    def describe(self) -> str: ...

    @property
    def transport_healthy(self) -> bool:
        """
        Whether the underlying transport stream is healthy
        """
        return self.stream is not None and not self._transport_failed and not self._terminated

    @property
    def usable(self) -> bool:
        """
        Whether the connection is established and initial handshakes were
        performed without error
        """
        return self.transport_healthy and self._ready

    @property
    def reusable(self) -> bool:
        """
        Whether the connection can be reused

        This property should be tested before returning a connection
        from the connection pool for reuse. In scenarios such as after using a connection
        for receiving push messages there might still be unconsumed messages in the
        internal buffer.

        If the connection is not reusable, invalidate it with a call to :meth:`invalidate`
        and discard it
        """
        return self.usable and (
            self._push_message_buffer_in.statistics().current_buffer_used == 0
            and self._streamed_message_buffer_in.statistics().current_buffer_used == 0
        )

    def register_connect_callback(
        self,
        callback: (Callable[[Self], None] | Callable[[Self], Awaitable[None]]),
    ) -> None:
        """
        Registers a callback that will be executed after the initial handshake
        is performed and before the connection is marked usable for regular
        commands.

        .. caution:: Any exception raised by a connect callback will not be
           handled and will result in an unusable connection.
        """
        self._connect_callbacks.append(callback)

    async def _trigger_connect_callbacks(self: Self) -> None:
        for callback in self._connect_callbacks:
            task = callback(self)
            if inspect.isawaitable(task):
                await task

    @abstractmethod
    async def _connect(self) -> ByteStream:
        """
        Establish and return the underlying transport connection to the Redis server.
        """
        ...

    async def run(self, *, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        """
        Establish a connection to the redis server and initiate any post connect callbacks.

        .. note:: This method can only be called once for an :class:`~coredis.connection.BaseConnection`
           instance. Once the connection closes either due to cancellation or errors it should be
           discarded.

        """
        if self._task_group:
            raise RuntimeError("Connection cannot be reused")

        def handle_errors(error: BaseExceptionGroup) -> None:
            # TODO: change _last_error to use the whole exception group
            #  once python 3.10 support is dropped and the library
            #  consistently uses exception groups
            self._last_error = self._last_error or error.exceptions[-1]

        try:
            self.stream = await self._connect()
            self.statistics.connected()
            try:
                with catch({Exception: handle_errors}):
                    async with (
                        self.stream,
                        self._write_buffer_in,
                        self._write_buffer_out,
                        self._push_message_buffer_in,
                        self._push_message_buffer_out,
                        self._streamed_message_buffer_in,
                        self._streamed_message_buffer_out,
                        create_task_group() as self._task_group,
                    ):
                        self._task_group.start_soon(self._reader_task)
                        self._task_group.start_soon(self._writer_task)
                        # setup connection
                        await self.__perform_handshake()
                        # *CAUTION* (and also maybe *FIXME*).
                        # There should be no code executed that could raise an
                        # exception after this line and before the task is marked
                        # as started.
                        task_status.started()
            finally:
                disconnect_exc = self._last_error or ConnectionError("Connection lost!")
                self._parser.on_disconnect()
                while self._requests:
                    request = self._requests.popleft()
                    request.fail(disconnect_exc)
                self.stream = None
                if self._ready:
                    # If a connection had successfully been established (including handshake)
                    # errors should no longer be raised and it is the responsibility of the
                    # downstream to ensure that `is_connected` is tested before using a connection
                    if self._last_error:
                        logger.info("Connection closed unexpectedly!", exc_info=True)
                    self._ready = False
                else:
                    # If a RedisError (raised ourselves) has resulted in an exception
                    # before a connection was completely established raise that.
                    # This is due to known handshake errors.
                    if isinstance(self._last_error, RedisError):
                        raise self._last_error
                    else:
                        logger.exception("Connection attempt failed unexpectedly!")
                        raise ConnectionError(
                            f"Unable to establish a connection to {self.location}"
                        ) from self._last_error
        except Exception as connection_error:
            self._last_error = connection_error
            if isinstance(connection_error, RedisError):
                raise
            else:
                # Wrap any other errors with a ConnectionError so that upstreams (pools) can
                # handle them explicitly as being part of connection creation if they want.
                raise ConnectionError(
                    f"Unable to establish a connection to {self.location}"
                ) from connection_error

    def invalidate(self, reason: str | None = None) -> None:
        """
        Forcefully mark the connection as unusable and release any associated resources.

        Call this when the connection state can no longer be trusted — for example,
        after a cancelled or timed-out request, or when unconsumed messages remain
        in the buffer. An invalidated connection should never be returned to the pool.

        .. note:: Calling this method on an already disconnected connection is safe
           and has no side effects
        """
        self._terminated = True
        if self._task_group:
            self._task_group.cancel_scope.cancel(reason)

    async def _reader_task(self) -> None:
        """
        Listen on the socket and run the parser, completing pending requests in
        FIFO order.
        """
        while self.stream is not None:
            with fail_after(self._max_idle_time):
                try:
                    data = await self.stream.receive()
                    self.statistics.data_received(len(data))
                except (EndOfStream, ClosedResourceError, BrokenResourceError) as err:
                    self._transport_failed = True
                    raise ConnectionError("Connection lost while receiving response") from err
                except Exception:
                    self._transport_failed = True
                    raise
                # If we're receiving data without any inflight requests
                # (push messages) tjhese need to be limited to avoid each
                # connection resulting in a hot read loop without any guaranteed
                # associated downstream consuming at the same rate.
                if not self._requests:
                    async with self._processing_budget:
                        self._data_received(data)
                else:
                    self._data_received(data)

    def _data_received(self, data: bytes) -> None:
        self._parser.feed(data)
        decode = self._requests[0].decode if self._requests else self._decode_responses
        response = self._parser.parse(
            decode=decode,
            encoding=self._requests[0].encoding if self._requests else self._encoding,
        )
        while not isinstance(response, NotEnoughData):
            if response[0] == DataType.PUSH:
                self._push_message_buffer_in.send_nowait(response[1])
            else:
                if self._requests:
                    request = self._requests.popleft()
                    if request.raise_exceptions and isinstance(response[1], RedisError):
                        request.fail(response[1])
                    else:
                        request.resolve(response[1])
                else:
                    self._streamed_message_buffer_in.send_nowait(response[1])
            decode = self._requests[0].decode if self._requests else self._decode_responses
            response = self._parser.parse(
                decode=decode,
                encoding=self._requests[0].encoding if self._requests else self._encoding,
            )

    async def _writer_task(self) -> None:
        """
        Continually empty the buffer and send the data to the server.
        """
        while self.stream is not None:
            requests = await self._write_buffer_out.receive()
            while self._write_buffer_out.statistics().current_buffer_used > 0:
                requests.extend(self._write_buffer_out.receive_nowait())
                await checkpoint()
            data = b"".join(
                self._packer.pack_commands(
                    [(request.command, *request.args) for request in requests]
                )
            )
            try:
                await self.stream.send(data)
                self.statistics.data_sent(len(data))
            except (ClosedResourceError, BrokenResourceError) as err:
                self._transport_failed = True
                raise ConnectionError("Connection lost while sending request") from err
            except Exception:
                self._transport_failed = True
                raise

    async def update_tracking_client(self, enabled: bool, client_id: int | None = None) -> bool:
        """
        Associate this connection to :paramref:`client_id` to
        relay any tracking notifications to.
        """
        try:
            params: list[RedisValueT] = (
                [b"ON", b"REDIRECT", client_id] if (enabled and client_id is not None) else [b"OFF"]
            )

            if await self.create_request(b"CLIENT TRACKING", *params, decode=False) != b"OK":
                raise ConnectionError("Unable to toggle client tracking")
            self.tracking_client_id = client_id
            return True
        except UnknownCommandError:  # noqa
            raise
        except Exception:  # noqa
            return False

    async def __perform_handshake(self) -> None:
        try:
            hello_command_args: list[int | str | bytes] = [3]
            if creds := (
                await self._credential_provider.get_credentials()
                if self._credential_provider
                else (
                    await UserPassCredentialProvider(
                        self._username, self._password
                    ).get_credentials()
                    if (self._username or self._password)
                    else None
                )
            ):
                hello_command_args.extend(
                    [
                        "AUTH",
                        creds.username,
                        creds.password or b"",
                    ]
                )
            hello_resp = await self.create_request(b"HELLO", *hello_command_args, decode=False)
            assert isinstance(hello_resp, (list, dict))
            resp3 = cast(dict[bytes, RedisValueT], hello_resp)
            assert resp3[b"proto"] == 3
            self.server_version = nativestr(resp3[b"version"])
            self.client_id = int(resp3[b"id"])
            if self.server_version >= "7.2":
                await self.create_request(
                    b"CLIENT SETINFO",
                    b"LIB-NAME",
                    b"coredis",
                )
                await self.create_request(
                    b"CLIENT SETINFO",
                    b"LIB-VER",
                    coredis.__version__,
                )

            if self._db:
                if await self.create_request(b"SELECT", self._db, decode=False) != b"OK":
                    raise ConnectionError(f"Invalid Database {self._db}")

            if self.client_name is not None:
                if (
                    await self.create_request(b"CLIENT SETNAME", self.client_name, decode=False)
                    != b"OK"
                ):
                    raise ConnectionError(f"Failed to set client name: {self.client_name}")

            if self._noevict:
                await self.create_request(b"CLIENT NO-EVICT", b"ON")

            if self._notouch:
                await self.create_request(b"CLIENT NO-TOUCH", b"ON")

            if self._noreply:
                await self.create_request(b"CLIENT REPLY", b"OFF", noreply=True)
                self._noreply_set = True
            await self._trigger_connect_callbacks()
            self._ready = True
        finally:
            self._handshake_attempted = True

    @property
    @_ensure_usable
    async def push_messages(self) -> AsyncGenerator[list[ResponseType], None]:
        """
        Generator to retrieve RESP3 push type messages sent by the server.
        The generator will yield each message received in order until the
        connection is lost and will raise an :exc:`~coredis.exceptions.ConnectionError`
        to signal that the generator has ended.
        """
        try:
            while True:
                yield await self._push_message_buffer_out.receive()
        except (EndOfStream, BrokenResourceError, ClosedResourceError) as err:
            raise ConnectionError("Connection lost while waiting for push messages") from err

    @property
    @_ensure_usable
    async def streamed_messages(self) -> AsyncGenerator[ResponseType, None]:
        """
        Generator to retrieve messages received from the server that were not
        sent as a response to a request made through this connection.
        The generator will yield each message received in order until the
        connection is lost and will raise an :exc:`~coredis.exceptions.ConnectionError`
        to signal that the generator has ended.
        """
        try:
            while True:
                yield await self._streamed_message_buffer_out.receive()
        except (EndOfStream, BrokenResourceError, ClosedResourceError) as err:
            raise ConnectionError("Connection lost") from err

    @_ensure_usable
    def send_command(
        self,
        command: bytes,
        *args: RedisValueT,
    ) -> None:
        """
        Queue a command to send to the server without
        associating a request to it. At the moment this is only useful in
        pubsub scenarios where commands such as ``SUBSCRIBE``, ``UNSUBSCRIBE``
        etc do not result in a response.
        """
        self._write_buffer_in.send_nowait(
            [
                Request(
                    connection=self,
                    command=command,
                    args=args,
                    decode=False,
                    expects_response=False,
                )
            ]
        )

    @_ensure_usable
    def create_request(
        self,
        command: bytes,
        *args: RedisValueT,
        noreply: bool | None = None,
        decode: RedisValueT | None = None,
        encoding: str | None = None,
        raise_exceptions: bool = True,
        timeout: float | None = None,
        disconnect_on_cancellation: bool = False,
    ) -> Request:
        """
        Queue a command to send to the server and create an
        associated request that can awaited for the response.
        """
        requests = []
        request_timeout: float | None = timeout or self._stream_timeout
        expects_response = not (self._noreply_set or noreply)
        if noreply and not self._noreply:
            requests.append(
                Request(
                    connection=self,
                    command=CommandName.CLIENT_REPLY,
                    args=(PureToken.SKIP,),
                    decode=bool(decode) if decode is not None else self._decode_responses,
                    encoding=encoding or self._encoding,
                    raise_exceptions=raise_exceptions,
                    expects_response=False,
                    disconnect_on_cancellation=disconnect_on_cancellation,
                )
            )
        request = Request(
            connection=self,
            command=command,
            args=args,
            decode=bool(decode) if decode is not None else self._decode_responses,
            encoding=encoding or self._encoding,
            raise_exceptions=raise_exceptions,
            response_timeout=request_timeout,
            expects_response=expects_response,
            disconnect_on_cancellation=disconnect_on_cancellation,
        )
        requests.append(request)
        self._write_buffer_in.send_nowait(requests)
        if expects_response:
            self._requests.append(request)
        return request

    @_ensure_usable
    def create_requests(
        self,
        commands: list[CommandInvocation],
        raise_exceptions: bool = True,
        timeout: float | None = None,
        disconnect_on_cancellation: bool = False,
    ) -> list[Request]:
        """
        Queue multiple commands to send to the server
        """
        request_timeout: float | None = timeout or self._stream_timeout
        requests = [
            Request(
                connection=self,
                command=cmd.command,
                args=cmd.args,
                decode=bool(cmd.decode) if cmd.decode is not None else self._decode_responses,
                encoding=cmd.encoding or self._encoding,
                raise_exceptions=raise_exceptions,
                response_timeout=request_timeout,
                disconnect_on_cancellation=disconnect_on_cancellation,
            )
            for cmd in commands
        ]
        self._write_buffer_in.send_nowait(requests)
        self._requests.extend(requests)
        return requests



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Hashable
from io import BytesIO
from typing import cast

from coredis.constants.resp import SYM_CRLF, SYM_TRUE, DataType
from coredis.exceptions import (
    AskError,
    AuthenticationFailureError,
    AuthenticationRequiredError,
    AuthorizationError,
    BusyLoadingError,
    ClusterCrossSlotError,
    ClusterDownError,
    ConnectionError,
    ExecAbortError,
    InvalidResponse,
    MovedError,
    NoKeyError,
    NoScriptError,
    NotBusyError,
    ProtocolError,
    ReadOnlyError,
    RedisError,
    ResponseError,
    StreamConsumerGroupError,
    StreamDuplicateConsumerGroupError,
    TryAgainError,
    UnblockedError,
    UnknownCommandError,
    WrongTypeError,
)
from coredis.typing import (
    Final,
    Literal,
    MutableSet,
    ResponsePrimitive,
    ResponseType,
    StringT,
)


class NotEnoughData:
    pass


NOT_ENOUGH_DATA: Final[NotEnoughData] = NotEnoughData()


class RESPNode:
    __slots__ = ("depth", "key", "node_type")
    depth: int
    key: Hashable
    node_type: Literal[DataType.PUSH, DataType.ARRAY, DataType.MAP, DataType.SET]

    def __init__(
        self,
        depth: int,
        node_type: Literal[DataType.PUSH, DataType.ARRAY, DataType.MAP, DataType.SET],
        key: (ResponsePrimitive | tuple[ResponsePrimitive, ...] | frozenset[ResponsePrimitive]),
    ):
        self.depth = depth
        self.node_type = node_type
        self.key = key

    @abstractmethod
    def append(self, item: ResponseType) -> None: ...

    def ensure_hashable(self, item: ResponseType) -> Hashable:
        if isinstance(item, (int, float, bool, str, bytes)):
            return item
        elif isinstance(item, set):
            return frozenset(self.ensure_hashable(cast(ResponseType, i)) for i in item)
        elif isinstance(item, list):
            return tuple(self.ensure_hashable(i) for i in item)
        elif isinstance(item, dict):
            return tuple(
                (cast(ResponsePrimitive, k), self.ensure_hashable(v)) for k, v in item.items()
            )
        return item


class ListNode(RESPNode):
    __slots__ = ("container",)

    def __init__(self, depth: int, node_type: Literal[DataType.PUSH, DataType.ARRAY]) -> None:
        self.container: list[ResponseType] = []
        super().__init__(depth, node_type, None)

    def append(self, item: ResponseType) -> None:
        self.depth -= 1
        self.container.append(item)


class DictNode(RESPNode):
    __slots__ = ("container", "key")

    def __init__(self, depth: int) -> None:
        self.container: dict[Hashable, ResponseType] = {}
        super().__init__(depth * 2, DataType.MAP, None)

    def append(self, item: ResponseType) -> None:
        self.depth -= 1
        if not self.key:
            self.key = self.ensure_hashable(item)
        else:
            self.container[self.key] = item
            self.key = None


class SetNode(RESPNode):
    __slots__ = ("container",)

    def __init__(self, depth: int) -> None:
        self.container: MutableSet[Hashable] = set()
        super().__init__(depth, DataType.SET, None)

    def append(
        self,
        item: ResponseType,
    ) -> None:
        self.depth -= 1
        self.container.add(self.ensure_hashable(item))


RESPScalar = (
    tuple[Literal[DataType.SIMPLE_STRING], StringT]
    | tuple[Literal[DataType.BULK_STRING], StringT | None]
    | tuple[Literal[DataType.VERBATIM], StringT]
    | tuple[Literal[DataType.INT], int]
    | tuple[Literal[DataType.BIGNUMBER], int]
    | tuple[Literal[DataType.DOUBLE], float]
    | tuple[Literal[DataType.BOOLEAN], bool]
    | tuple[Literal[DataType.NONE], None]
    | tuple[Literal[DataType.ERROR], RedisError]
)
RESPContainer = (
    tuple[Literal[DataType.ARRAY], list[ResponseType]]
    | tuple[Literal[DataType.PUSH], list[ResponseType]]
    | tuple[Literal[DataType.MAP], dict[Hashable, ResponseType]]
    | tuple[Literal[DataType.SET], MutableSet[Hashable]]
)

UnpackedResponse = RESPScalar | RESPContainer


class Parser:
    """ """

    EXCEPTION_CLASSES: dict[str, type[RedisError] | dict[str, type[RedisError]]] = {
        "ASK": AskError,
        "BUSYGROUP": StreamDuplicateConsumerGroupError,
        "CLUSTERDOWN": ClusterDownError,
        "CROSSSLOT": ClusterCrossSlotError,
        "ERR": {
            "max number of clients reached": ConnectionError,
            "unknown command": UnknownCommandError,
            "unknown subcommand": UnknownCommandError,
            "sub command not supported": UnknownCommandError,
        },
        "EXECABORT": ExecAbortError,
        "LOADING": BusyLoadingError,
        "NOSCRIPT": NoScriptError,
        "MOVED": MovedError,
        "NOAUTH": AuthenticationRequiredError,
        "NOGROUP": StreamConsumerGroupError,
        "NOKEY": NoKeyError,
        "NOPERM": AuthorizationError,
        "NOPROTO": ProtocolError,
        "NOTBUSY": NotBusyError,
        "READONLY": ReadOnlyError,
        "TRYAGAIN": TryAgainError,
        "UNBLOCKED": UnblockedError,
        "WRONGPASS": AuthenticationFailureError,
        "WRONGTYPE": WrongTypeError,
    }

    def __init__(self) -> None:
        self.localbuffer: BytesIO = BytesIO(b"")
        self.bytes_read: int = 0
        self.bytes_written: int = 0
        self.nodes: list[ListNode | SetNode | DictNode] = []

    def feed(self, data: bytes) -> None:
        self.localbuffer.seek(self.bytes_written)
        self.bytes_written += self.localbuffer.write(data)
        self.localbuffer.seek(self.bytes_read)

    def on_disconnect(self) -> None:
        """Called when the stream disconnects"""
        if not self.localbuffer.closed:
            self.localbuffer.seek(0)
            self.localbuffer.truncate()
            self.bytes_read = self.bytes_written = 0
        self.nodes.clear()

    def can_read(self) -> bool:
        return (self.bytes_written - self.bytes_read) > 0

    def try_decode(self, data: bytes, encoding: str) -> bytes | str:
        try:
            return data.decode(encoding)
        except ValueError:
            return data

    def parse(
        self,
        decode: bool = False,
        encoding: str | None = None,
    ) -> UnpackedResponse | NotEnoughData:
        """
        :param decode: Whether to decode simple or bulk strings
        :param encoding: The encoding to use if :paramref:`decode` is set
        :return: The next available parsed response read from the connection.
         If there is not enough data on the wire a ``NotEnoughData`` instance
         will be returned.
        """
        while True:
            data = self.localbuffer.readline()
            if not data[-2::] == SYM_CRLF:
                return NOT_ENOUGH_DATA
            data_len = len(data)
            self.bytes_read += data_len
            marker, chunk = data[0], data[1:-2]
            response: ResponseType = None
            match marker:
                case DataType.SIMPLE_STRING:
                    response = chunk
                    if decode and encoding:
                        response = self.try_decode(response, encoding)
                case DataType.BULK_STRING | DataType.VERBATIM:
                    length = int(chunk)
                    if length == -1:
                        response = None
                    else:
                        if (self.bytes_written - self.bytes_read) < length + 2:
                            self.bytes_read -= data_len
                            return NOT_ENOUGH_DATA
                        data = self.localbuffer.read(length + 2)
                        self.bytes_read += length + 2
                        response = data[:-2]
                        if marker == DataType.VERBATIM:
                            if response[:3] != b"txt":
                                raise InvalidResponse(
                                    f"Unexpected verbatim string of type {response[:3]!r}"
                                )
                            response = response[4:]
                        if decode and encoding:
                            response = self.try_decode(response, encoding)
                case DataType.INT | DataType.BIGNUMBER:
                    response = int(chunk)
                case DataType.DOUBLE:
                    response = float(chunk)
                case DataType.NONE:
                    response = None
                case DataType.BOOLEAN:
                    response = chunk[0] == ord(SYM_TRUE)
                case DataType.ARRAY | DataType.PUSH | DataType.MAP | DataType.SET:
                    length = int(chunk)
                    if length >= 0:
                        match marker:
                            case DataType.ARRAY:
                                self.nodes.append(ListNode(length, DataType.ARRAY))
                            case DataType.PUSH:
                                self.nodes.append(ListNode(length, DataType.PUSH))
                            case DataType.MAP:
                                self.nodes.append(DictNode(length))
                            case DataType.SET:
                                self.nodes.append(SetNode(length))
                        if length > 0:
                            continue
                case DataType.ERROR:
                    response = self.parse_error(bytes(chunk).decode())
                case _:
                    raise InvalidResponse(
                        f"Protocol Error: Unknown RESP data type: {chr(marker)!r}"
                    )
            if self.nodes:
                if self.nodes[-1].depth > 0:
                    self.nodes[-1].append(response)
                    if self.nodes[-1].depth > 1:
                        continue

                while len(self.nodes) > 1 and self.nodes[-1].depth == 0:
                    self.nodes[-2].append(self.nodes[-1].container)
                    self.nodes.pop()

            if len(self.nodes) == 1 and self.nodes[-1].depth == 0:
                node = self.nodes.pop()
                parsed = cast(UnpackedResponse, (DataType(node.node_type), node.container))
                break
            if not self.nodes:
                parsed = cast(UnpackedResponse, (DataType(marker), response))
                break

        if self.bytes_read == self.bytes_written:
            self.localbuffer.seek(0)
            self.localbuffer.truncate()
            self.bytes_read = self.bytes_written = 0
        return parsed

    def parse_error(self, response: str) -> RedisError:
        """
        Parse an error response

        :meta private:
        """
        error_code = response.split(" ")[0]
        if error_code in self.EXCEPTION_CLASSES:
            response = response[len(error_code) + 1 :]
            exception_class = self.EXCEPTION_CLASSES[error_code]

            if isinstance(exception_class, dict):
                options = exception_class.items()
                exception_class = ResponseError
                for err, exc in options:
                    if response.lower().startswith(err):
                        exception_class = exc
                        break
            return exception_class(response)
        return ResponseError(response)



from __future__ import annotations

from coredis.typing import Unpack

from ._base import BaseConnectionParams
from ._tcp import TCPConnection, TCPLocation


class ClusterConnection(TCPConnection):
    "Manages TCP communication to and from a Redis Cluster node"

    def __init__(
        self,
        location: TCPLocation,
        *,
        read_from_replicas: bool = False,
        **kwargs: Unpack[BaseConnectionParams],
    ) -> None:
        self._read_from_replicas = read_from_replicas
        super().__init__(location, **kwargs)
        if self._read_from_replicas:
            self.register_connect_callback(self._set_readonly)

    async def _set_readonly(self, connection: ClusterConnection) -> None:
        """
        Read only cluster connections need to explicitly
        request readonly during handshake
        """
        if not (await connection.create_request(b"READONLY", decode=False)) == b"OK":
            raise ConnectionError("Could not set READONLY mode for cluster replica")

    def describe(self) -> str:
        return f"ClusterConnection<path={self.location.host},db={self.location.port}>"



from __future__ import annotations

from ._base import BaseConnectionPool, BaseConnectionPoolParams
from ._basic import ConnectionPool, ConnectionPoolParams
from ._cluster import ClusterConnectionPool, ClusterConnectionPoolParams
from ._sentinel import SentinelConnectionPool

__all__ = [
    "BaseConnectionPool",
    "BaseConnectionPoolParams",
    "ConnectionPool",
    "ConnectionPoolParams",
    "ClusterConnectionPool",
    "ClusterConnectionPoolParams",
    "SentinelConnectionPool",
]



from __future__ import annotations

from collections import ChainMap, OrderedDict
from typing import Any, cast

from coredis._json import json
from coredis._utils import EncodingInsensitiveDict
from coredis.modules.response.types import (
    HybridResult,
    SearchAggregationResult,
    SearchDocument,
    SearchResult,
)
from coredis.response._callbacks import ResponseCallback
from coredis.typing import (
    AnyStr,
    ResponsePrimitive,
    ResponseType,
    StringT,
)


class SearchConfigCallback(
    ResponseCallback[
        dict[AnyStr, ResponseType],
        dict[AnyStr, ResponsePrimitive],
    ]
):
    def transform(
        self,
        response: dict[AnyStr, ResponseType],
    ) -> dict[AnyStr, ResponsePrimitive]:
        config = {}
        for item, value in response.items():
            try:
                config[item] = json.loads(value)
            except (ValueError, TypeError):
                config[item] = value
        return config


class SearchResultCallback(
    ResponseCallback[
        dict[AnyStr, ResponseType],
        SearchResult[AnyStr],
    ]
):
    def transform(
        self,
        response: dict[AnyStr, ResponseType],
    ) -> SearchResult[AnyStr]:
        results = []
        search_results = EncodingInsensitiveDict(response)
        for result in search_results["results"]:
            result = EncodingInsensitiveDict(result)
            score_explain = None
            if self.options.get("explainscore"):
                score, score_explain = result.get("score")
            else:
                score = result.get("score", None)
            fields = EncodingInsensitiveDict(result.get("extra_attributes", {}))
            if "$" in fields:
                fields = json.loads(fields.pop("$"))
            results.append(
                SearchDocument(
                    result["id"],
                    float(score) if score else None,
                    score_explain,
                    result["payload"] if self.options.get("withpayloads") else None,
                    result["sortkey"] if self.options.get("withsortkeys") else None,
                    cast(dict[Any, ResponseType], fields),
                )
            )
        return SearchResult[AnyStr](search_results["total_results"], tuple(results))


class AggregationResultCallback(
    ResponseCallback[
        dict[StringT, ResponseType] | list[dict[StringT, ResponseType] | int],
        SearchAggregationResult[AnyStr],
    ]
):
    def transform(
        self, response: dict[StringT, ResponseType] | list[dict[StringT, ResponseType] | int]
    ) -> SearchAggregationResult[AnyStr]:
        if self.options.get("with_cursor"):
            agg, cursor = cast(tuple[dict[AnyStr, ResponseType], int], tuple(response))
        else:
            agg = cast(dict[AnyStr, ResponseType], response)
            cursor = None
        aggregations = EncodingInsensitiveDict(agg)
        return SearchAggregationResult[AnyStr](
            [
                {
                    r: self.try_json(self.options, v)
                    for r, v in EncodingInsensitiveDict(k)["extra_attributes"].items()
                }
                for k in aggregations["results"]
            ],
            cursor,
        )

    @staticmethod
    def try_json(options: dict[str, Any], value: StringT) -> ResponseType:
        if not options.get("dialect", None) == 3:
            return cast(ResponseType, value)
        try:
            return cast(ResponseType, json.loads(value))
        except ValueError:
            return cast(ResponseType, value)


class SpellCheckCallback(
    ResponseCallback[
        dict[StringT, ResponseType],
        dict[AnyStr, OrderedDict[AnyStr, float]],
    ]
):
    def transform(
        self,
        response: dict[StringT, ResponseType],
    ) -> dict[AnyStr, OrderedDict[AnyStr, float]]:
        corrections = EncodingInsensitiveDict(response)
        return {
            key: OrderedDict(ChainMap(*result)) for key, result in corrections["results"].items()
        }


class HybridSearchCallback(
    ResponseCallback[
        dict[StringT, int | float | list[StringT | dict[StringT, StringT]]], HybridResult[AnyStr]
    ]
):
    def transform(
        self,
        response: dict[StringT, int | float | list[StringT | dict[StringT, StringT]]],
    ) -> HybridResult[AnyStr]:
        results = EncodingInsensitiveDict(response)
        return HybridResult[AnyStr](
            cast(int, results["total_results"]),
            cast(float, results["execution_time"]),
            cast(list[AnyStr], results["warnings"]),
            cast(list[dict[AnyStr, AnyStr]], results["results"]),
        )



from __future__ import annotations

import warnings
from contextlib import asynccontextmanager
from typing import Any

from anyio import TASK_STATUS_IGNORED, fail_after, sleep
from anyio.abc import TaskStatus

from coredis._concurrency import Queue, QueueEmpty, QueueFull
from coredis._utils import logger, query_param_to_bool
from coredis.cluster._discovery import DiscoveryService
from coredis.cluster._layout import ClusterLayout
from coredis.cluster._node import ClusterNodeLocation
from coredis.connection import (
    BaseConnectionParams,
    ClusterConnection,
)
from coredis.connection._tcp import TCPLocation
from coredis.exceptions import RedisError
from coredis.patterns.cache import AbstractCache, ClusterTrackingCache
from coredis.pool._basic import ConnectionPoolParams
from coredis.typing import (
    AsyncGenerator,
    Callable,
    ClassVar,
    Iterable,
    Node,
    NotRequired,
    Self,
    Unpack,
)

from ._base import BaseConnectionPool


class ClusterConnectionPoolParams(ConnectionPoolParams[ClusterConnection]):
    """
    Parameters accepted by :class:`coredis.pool.ClusterConnectionPool`
    """

    #: The initial collection of nodes to use to map the cluster solts to individual primary & replica nodes.
    startup_nodes: NotRequired[Iterable[Node | TCPLocation]]
    #: Skips the check of cluster-require-full-coverage config, useful for clusters
    #: without the :rediscommand:`CONFIG` command (For example with AWS Elasticache)
    skip_full_coverage_check: NotRequired[bool]
    #: Whether to use the value of ``max_connections``
    #: on a per node basis or cluster wide. If ``False``  the per-node connection pools will have
    #: a maximum size of :paramref:`max_connections` divided by the number of nodes in the cluster.
    max_connections_per_node: NotRequired[bool]
    #: If ``True`` connections to replicas will be returned for readonly commands
    read_from_replicas: NotRequired[bool]
    #: Interval (in seconds) for performing a cleanup of the pool to
    #: remove any connections that are no longer in the cluster layout.
    gc_interval: NotRequired[int]


class ClusterConnectionPool(BaseConnectionPool[ClusterConnection]):
    """
    Custom connection pool for :class:`~coredis.RedisCluster` client
    """

    URL_QUERY_ARGUMENT_PARSERS: ClassVar[
        dict[str, Callable[..., int | float | bool | str | None]]
    ] = {
        **BaseConnectionPool.URL_QUERY_ARGUMENT_PARSERS,
        "max_connections_per_node": query_param_to_bool,
        "reinitialize_steps": int,
        "skip_full_coverage_check": query_param_to_bool,
        "read_from_replicas": query_param_to_bool,
        "gc_interval": int,
    }

    connection_class: type[ClusterConnection]

    _cluster_available_connections: dict[TCPLocation, Queue[ClusterConnection]]
    _online_connections: set[ClusterConnection]

    def __init__(
        self,
        startup_nodes: Iterable[Node | TCPLocation],
        *,
        connection_class: type[ClusterConnection] = ClusterConnection,
        max_connections: int | None = None,
        max_connections_per_node: bool = False,
        reinitialize_steps: int | None = None,
        skip_full_coverage_check: bool = False,
        nodemanager_follow_cluster: bool = True,
        readonly: bool = False,
        read_from_replicas: bool = False,
        timeout: float | None = None,
        gc_interval: int = 30,
        _cache: AbstractCache | None = None,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ):
        """
        Cluster aware connection pool that tracks and manages sub pools for each node
        in the redis cluster

        Changes
          - .. deprecated:: 6.2.0

            - :paramref:`startup_nodes` should not be passed as dictionaries
               and instead migrate to use instances of :class:`~coredis.connection.TCPLocation`

          - .. versionchanged:: 4.4.0

            - :paramref:`nodemanager_follow_cluster` now defaults to ``True``

          - .. deprecated:: 4.4.0

            - :paramref:`readonly` renamed to :paramref:`read_from_replicas`

        :param startup_nodes: The initial collection of nodes to use to map the cluster
         solts to individual primary & replica nodes.
        :param connection_class: The connection class to use when creating new connections
        :param max_connections: Maximum number of connections to allow concurrently from this
         client. If the value is ``None`` it will default to 64.
        :param max_connections_per_node: Whether to use the value of :paramref:`max_connections`
         on a per node basis or cluster wide. If ``False``  the per-node connection pools will have
         a maximum size of :paramref:`max_connections` divided by the number of nodes in the cluster.
        :param timeout: Number of seconds to block when trying to obtain a connection.
        :param skip_full_coverage_check:
         Skips the check of cluster-require-full-coverage config, useful for clusters
         without the :rediscommand:`CONFIG` command (For example with AWS Elasticache)
        :param nodemanager_follow_cluster: The node manager will during initialization try
         the last set of nodes that it was operating on. This will allow the client to drift
         along side the cluster if the cluster nodes move around alot.
        :param read_from_replicas: If ``True`` connections to replicas will be returned for readonly
         commands
        :param gc_interval: Interval (in seconds) for performing a cleanup of the pool to
         remove any connections that are no longer in the cluster layout.
        :param connection_kwargs: arguments to pass to the :paramref:`connection_class`
         constructor when creating a new connection
        """
        super().__init__(
            connection_class=connection_class,
            **connection_kwargs,
        )
        self._initialized = False
        self._gc_interval = gc_interval
        self.timeout = timeout
        self.max_connections = max_connections or 64
        self.max_connections_per_node = max_connections_per_node
        # TODO: Remove support for Node
        if any(isinstance(node, dict) for node in startup_nodes):
            warnings.warn(
                "Use coredis.connection.TCPLocation to specify startup nodes",
                DeprecationWarning,
                stacklevel=2,
            )

        self.startup_nodes = [
            node if isinstance(node, TCPLocation) else TCPLocation(node["host"], node["port"])
            for node in startup_nodes
        ]
        self.cluster_layout = ClusterLayout(
            DiscoveryService(
                startup_nodes=self.startup_nodes,
                skip_full_coverage_check=skip_full_coverage_check,
                follow_cluster=nodemanager_follow_cluster,
                **connection_kwargs,
            ),
            error_threshold=reinitialize_steps or 2,
        )
        self.connection_kwargs = connection_kwargs
        self.read_from_replicas = read_from_replicas or readonly
        # TODO: Use the `max_failures` argument of tracking cache
        self.cache = ClusterTrackingCache(self, _cache) if _cache else None
        self._reset()

        if "stream_timeout" not in self.connection_kwargs:
            self.connection_kwargs["stream_timeout"] = None

    @classmethod
    def from_url(
        cls: type[Self],
        url: str,
        *,
        decode_components: bool = False,
        **kwargs: Unpack[ClusterConnectionPoolParams],
    ) -> Self:
        """
        Returns a cluster connection pool configured from the given URL.
        """
        location, merged_options = cls._parse_url(
            url, decode_components, kwargs, ClusterConnectionPoolParams
        )
        if "startup_nodes" not in merged_options and location:
            assert isinstance(location, TCPLocation)
            merged_options["startup_nodes"] = [Node(host=location.host, port=location.port)]
        merged_options["connection_class"] = ClusterConnection

        return cls(
            **merged_options,
        )

    def __repr__(self) -> str:
        """
        Returns a string with all unique ip:port combinations that this pool
        is connected to
        """

        return "{}<{}>".format(
            type(self).__name__,
            ", ".join([node.name for node in list(self.cluster_layout.nodes)]),
        )

    async def _initialize(self) -> None:
        await self.cluster_layout.initialize()
        total_nodes = len(self.cluster_layout.nodes)
        if not self.max_connections_per_node and self.max_connections < total_nodes:
            warnings.warn(
                f"The value of max_connections={self.max_connections} "
                "should be atleast equal to the number of nodes "
                f"({total_nodes}) in the cluster and has been increased by "
                f"{total_nodes - self.max_connections} connections."
            )
            self.max_connections = total_nodes
        await self.task_group.start(self.cluster_layout.monitor)
        await self.task_group.start(self._cleanup)
        if self.cache:
            # TODO: handle cache failure so that the pool doesn't die
            #  if the cache fails.
            await self.task_group.start(self.cache.run)

    async def get_connection(
        self, node: ClusterNodeLocation | None = None, primary: bool = True, **options: Any
    ) -> ClusterConnection:
        """
        Acquires a connection from the cluster pool. If no node
        is specified a random node is picked. The connection
        must be returned back to the pool using the :meth:`release`
        method.

        :param node:  The node for which to get a connection from
        :param primary: If False a connection from the replica will be returned
        """
        connection: ClusterConnection
        if node:
            connection = await self._get_connection_by_node(node)
        else:
            connection = await self._get_random_connection(primary=primary)
        return connection

    @asynccontextmanager
    async def acquire(
        self,
        node: ClusterNodeLocation | None = None,
        primary: bool = True,
        **options: Any,
    ) -> AsyncGenerator[ClusterConnection]:
        """
        Acquires a connection from the cluster pool. If no node
        is specified a random node is picked. The connection
        will be automatically released back to the pool when
        the context manager exits.

        :param node:  The node for which to get a connection from
        :param primary: If False a connection from the replica will be returned
        """
        connection = await self.get_connection(node=node, primary=primary, **options)
        yield connection
        self.release(connection)

    def release(self, connection: ClusterConnection) -> None:
        """Releases the connection back to the pool"""
        assert isinstance(connection, ClusterConnection)
        try:
            if connection.usable:
                self._node_pool(connection.location).put_nowait(connection)
        except QueueFull:
            pass

    def disconnect(self) -> None:
        for connection in self._online_connections:
            connection.invalidate()
        self._online_connections.clear()

    def _reset(self) -> None:
        """Resets the connection pool back to a clean state"""
        self._cluster_available_connections = {}
        self._online_connections = set()

    async def _wrap_connection(
        self,
        connection: ClusterConnection,
        *,
        task_status: TaskStatus[None | Exception] = TASK_STATUS_IGNORED,
    ) -> None:
        try:
            await connection.run(task_status=task_status)
        except RedisError as error:
            # Only coredis.exception.RedisError is explictly caught and returned with the task status
            # As these are clear signals that an error case was handled by the connection
            task_status.started(error)
        finally:
            node_pool = self._node_pool(connection.location)
            self._online_connections.discard(connection)
            if connection in node_pool:
                node_pool.remove(connection)
                node_pool.append_nowait(None)

    async def _make_node_connection(self, node: ClusterNodeLocation) -> ClusterConnection:
        """Creates a new connection to a node"""

        location = TCPLocation(node.host, node.port)
        connection = self.connection_class(
            location=location,
            read_from_replicas=self.read_from_replicas and node.server_type == "replica",
            **self.connection_kwargs,
        )
        if err := await self.task_group.start(self._wrap_connection, connection):
            raise err
        self._online_connections.add(connection)
        return connection

    def _node_pool(self, location: TCPLocation) -> Queue[ClusterConnection]:
        if self._cluster_available_connections.get(location) is None:
            self._cluster_available_connections[location] = self._default_node_queue()
        return self._cluster_available_connections[location]

    def _default_node_queue(
        self,
    ) -> Queue[ClusterConnection]:
        q_size = max(
            1,
            self.max_connections
            if self.max_connections_per_node
            else self.max_connections // len(list(self.cluster_layout.nodes)),
        )
        return Queue[ClusterConnection](q_size)

    async def _get_random_connection(self, primary: bool = False) -> ClusterConnection:
        return await self._get_connection_by_node(self.cluster_layout.random_node(primary))

    async def _get_connection_by_node(self, node: ClusterNodeLocation) -> ClusterConnection:
        location = TCPLocation(node.host, node.port)
        with fail_after(self.timeout):
            connection = await self._node_pool(location).get()

        if not connection or not connection.reusable:
            # If the connection was in the pool but is "dirty" it should be
            # invalidated (to avoid leaking) and discarded.
            if connection:
                connection.invalidate()
            try:
                connection = await self._make_node_connection(node)
            except Exception:
                self._node_pool(location).append_nowait(None)
                raise

        return connection

    async def _cleanup(self, task_status: TaskStatus[None] = TASK_STATUS_IGNORED) -> None:
        task_status.started()
        while True:
            if self.cluster_layout.nodes:
                for location, pool in list(self._cluster_available_connections.items()):
                    if not self.cluster_layout.node_for_location(location):
                        dead_queue = self._cluster_available_connections.pop(location)
                        connections = []
                        try:
                            while True:
                                if c := dead_queue.get_nowait():
                                    connections.append(c)
                        except QueueEmpty:
                            pass
                        logger.info(
                            f"Node for {location} is no longer in cluster layout, releasing from connection pool (connections: {len(connections)}"
                        )
                        for connection in connections:
                            connection.invalidate()
            await sleep(self._gc_interval)



from __future__ import annotations

import dataclasses
import math
import time


@dataclasses.dataclass
class ConnectionStatistics:
    """
    Connection Statistics
    """

    created_at: float = dataclasses.field(init=False, default_factory=lambda: time.perf_counter())
    connected_at: float = dataclasses.field(init=False, default=-math.inf)
    first_byte_received_at: float = dataclasses.field(init=False, default=-math.inf)
    last_byte_received_at: float = dataclasses.field(init=False, default=-math.inf)
    last_byte_written_at: float = dataclasses.field(init=False, default=-math.inf)
    last_request_created_at: float = dataclasses.field(init=False, default=-math.inf)
    last_request_resolved_at: float = dataclasses.field(init=False, default=-math.inf)
    bytes_read: int = 0
    bytes_written: int = 0
    requests_created: int = 0
    requests_resolved: int = 0

    @property
    def rtt(self) -> float:
        """
        Approximate RTT to the node, measured by connection
        establishment duration.
        """
        if math.isinf(self.connected_at):
            return math.nan
        return self.connected_at - self.created_at

    @property
    def ttfb(self) -> float:
        """
        Time to receive first byte since the connection was established
        """
        if math.isinf(self.first_byte_received_at):
            return math.nan
        return self.first_byte_received_at - self.connected_at

    @property
    def requests_pending(self) -> float:
        """
        Total requests still awaiting a response to be resolved
        """
        return self.requests_created - self.requests_resolved

    def connected(self) -> None:
        if math.isinf(self.connected_at):
            self.connected_at = time.perf_counter()

    def data_received(self, num_bytes: int) -> None:
        now = time.perf_counter()
        if math.isinf(self.first_byte_received_at):
            self.first_byte_received_at = now
        self.last_byte_received_at = now
        self.bytes_read += num_bytes

    def data_sent(self, num_bytes: int) -> None:
        now = time.perf_counter()
        self.bytes_written += num_bytes
        self.last_byte_written_at = now

    def request_created(self) -> None:
        now = time.perf_counter()
        self.last_request_created_at = now
        self.requests_created += 1

    def request_resolved(self) -> None:
        now = time.perf_counter()
        self.last_request_resolved_at = now
        self.requests_resolved += 1



from __future__ import annotations

import copy
import functools
from types import GenericAlias
from typing import Any, cast, get_origin

from coredis._protocols import AbstractExecutor
from coredis.retry import RetryPolicy
from coredis.typing import (
    Awaitable,
    Callable,
    ExecutionParameters,
    Generator,
    Serializable,
    TypeAdapter,
    TypeIs,
    TypeVar,
    ValueT,
)

#: Covariant type used for generalizing :class:`~coredis.command.CommandRequest`
CommandResponseT = TypeVar("CommandResponseT", covariant=True)

TransformedResponse = TypeVar("TransformedResponse")
empty_adapter = TypeAdapter()


def is_type_like(obj: object) -> TypeIs[type[Any]]:
    """
    Return True if ``obj`` is type-like and should be treated as a
    deserialization target rather than a callable transformer.
    """
    return isinstance(obj, type) or isinstance(obj, GenericAlias) or get_origin(obj) is not None


class CommandRequest(Awaitable[CommandResponseT]):
    _response: Awaitable[CommandResponseT]

    def __init__(
        self,
        client: AbstractExecutor,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., CommandResponseT],
        execution_parameters: ExecutionParameters | None = None,
        **kwargs: Any,
    ) -> None:
        """
        The default command request object which is returned by all
        methods mirroring redis commands.

        :param client: The instance of the :class:`coredis.Redis` that
         will be used to call :meth:`~coredis.Redis.execute_command`
        :param name:  The name of the command
        :param arguments:  All arguments (in redis format) to be passed to the command
        :param callback: The callback to be used to transform the RESP response
        :param execution_parameters:  Any additional parameters to be passed to
         :meth:`coredis.Redis.execute_command`
        """
        self.name = name
        self.callback = callback
        self.execution_parameters = execution_parameters or {}
        self.client: AbstractExecutor = client
        self.arguments = tuple(
            self.type_adapter.serialize(k) if isinstance(k, Serializable) else k for k in arguments
        )
        self.kwargs = kwargs

    def run(self) -> Awaitable[CommandResponseT]:
        if not hasattr(self, "_response"):
            self._response = self.client.execute_command(
                self, self.callback, **self.execution_parameters
            )

        return self._response

    def retry(
        self,
        policy: RetryPolicy,
        failure_hook: Callable[..., Awaitable[Any]] | None = None,
    ) -> CommandRequest[CommandResponseT]:
        """

        :param policy: Retry policy to use
        :param failure_hook: Callable[..., Awaitable[Any]]
                      | dict[type[BaseException], Callable[..., Awaitable[None]]]
                      | None = None,
        :return: A retryable version of the command object

        Calling ``retry`` is essentially the same as explicitly using
        a retry policy when calling a redis command. For example the following two
        examples that try to push to a list upto 2 times if a :exc:`~coredis.exceptions.RedisError`
        is encountered, are equivalent.

        With :meth:`retry`::

            async with coredis.Redis() as client:
                await client.lpush("mylist", [1,2,3]).retryable(
                    coredis.retry.ConstantRetryPolicy(
                        (coredis.exceptions.RedisError,), 2, 1
                    ),
                )

        Explicitly retrying::

            retry_policy = coredis.retry.ConstantRetryPolicy(
                (coredis.exceptions.RedisError,), 2, 1
            )

            async with coredis.Redis() as client:
                await retry_policy.call_with_retries(lambda:client.lpush("mylist", [1,2,3])

        """
        return RetryableCommandRequest(
            self.client,
            self.name,
            *self.arguments,
            callback=self.callback,
            execution_parameters=self.execution_parameters,
            policy=policy,
            failure_hook=failure_hook,
        )

    def transform(
        self,
        transformer: type[TransformedResponse] | Callable[[CommandResponseT], TransformedResponse],
    ) -> CommandRequest[TransformedResponse]:
        """
        :param transformer: A type that was registered with the client
         using :meth:`~coredis.typing.TypeAdapter.register_deserializer`
         or decorated by :meth:`~coredis.typing.TypeAdapter.deserializer`
         or a callable that takes a single argument (the original response)
         and returns the transformed response.

        :return: a command request object that when awaited will return the
         transformed response

         For example when used with a redis command::

           client = coredis.Redis(....)
           @client.type_adapter.deserializer
           def _(value: bytes) -> int:
               return int(value)

           async with client:
               await client.set("fubar", 1)
               raw: bytes = await client.get("fubar")
               int_value: int = await client.get("fubar").transform(int)
               float_value: float = await client.get("fubar").transform(lambda value: float(value))
        """
        return TransformedCommandRequest(
            self,
            transformer=transformer,
        )

    @property
    def type_adapter(self) -> TypeAdapter:
        from coredis.client import Client

        if isinstance(self.client, Client):
            return self.client.type_adapter

        return empty_adapter

    def __await__(self) -> Generator[Any, Any, CommandResponseT]:
        return self.run().__await__()


class TransformedCommandRequest(CommandRequest[TransformedResponse]):
    def __init__(
        self,
        parent: CommandRequest[CommandResponseT],
        transformer: type[TransformedResponse] | Callable[[CommandResponseT], TransformedResponse],
    ) -> None:
        """
        A command request object that has a tranformed response based
        on the :paramref:`transformer` parameter.

        :param parent: The original command request that the transformer will
         be applied to.
        :param transformer: A type that was registered with the client
         using :meth:`~coredis.typing.TypeAdapter.register_deserializer`
         or decorated by :meth:`~coredis.typing.TypeAdapter.deserializer`
         or a callable that takes a single argument (the original response)
         and returns the transformed response.
        """
        self.transformer = transformer
        self.parent = parent
        self.client = parent.client
        self.transform_func = cast(
            Callable[..., TransformedResponse],
            (
                functools.partial(self.type_adapter.deserialize, return_type=self.transformer)
                if is_type_like(self.transformer)
                else self.transformer
            ),
        )

    def __await__(self) -> Generator[Any, Any, TransformedResponse]:
        async def _parent() -> TransformedResponse:
            original = await self.parent
            return self.transform_func(original)

        return _parent().__await__()


class RetryableCommandRequest(CommandRequest[CommandResponseT]):
    def __init__(
        self,
        client: AbstractExecutor,
        name: bytes,
        *arguments: ValueT,
        callback: Callable[..., CommandResponseT],
        execution_parameters: ExecutionParameters | None = None,
        policy: RetryPolicy,
        failure_hook: Callable[..., Awaitable[Any]]
        | dict[type[BaseException], Callable[..., Awaitable[None]]]
        | None = None,
        **kwargs: Any,
    ) -> None:
        """
        A retryable command request object.

        :param client: The instance of the :class:`coredis.Redis` that
        will be used to call :meth:`~coredis.Redis.execute_command`
        :param name:  The name of the command
        :param arguments:  All arguments (in redis format) to be passed to the command
        :param callback: The callback to be used to transform the RESP response
        :param execution_parameters:  Any additional parameters to be passed to
        :meth:`coredis.Redis.execute_command`
        :param policy: The retry policy to use when executing the command
        :param failure_hook: if provided and is a callable it will be
         called everytime a retryable exception is encountered. If it is a mapping
         of exception types to callables, the first exception type that is a parent
         of any encountered exception will be called.
        """
        self.policy = policy
        self.failure_hook = failure_hook
        super().__init__(
            client,
            name,
            *arguments,
            callback=callback,
            execution_parameters=execution_parameters,
            policy=self.policy,
            failure_hook=self.failure_hook,
            **kwargs,
        )

    def __await__(self) -> Generator[Any, Any, CommandResponseT]:
        return self.policy.call_with_retries(
            lambda: copy.copy(self).run(), failure_hook=self.failure_hook
        ).__await__()



from __future__ import annotations

import warnings
from collections import defaultdict

from coredis.connection import BaseConnectionParams, TCPLocation
from coredis.exceptions import ConnectionError, RedisClusterError, RedisError, ResponseError
from coredis.typing import Final, Iterable, Unpack

from ._node import ClusterNodeLocation

HASH_SLOTS: Final = 16384
HASH_SLOTS_SET: Final = set(range(HASH_SLOTS))


class DiscoveryService:
    def __init__(
        self,
        startup_nodes: Iterable[TCPLocation] | None = None,
        skip_full_coverage_check: bool = False,
        follow_cluster: bool = True,
        **connection_kwargs: Unpack[BaseConnectionParams],
    ) -> None:
        self._connection_kwargs = connection_kwargs
        self._startup_nodes: list[ClusterNodeLocation] = (
            []
            if startup_nodes is None
            else list(ClusterNodeLocation(n.host, n.port) for n in startup_nodes if n)
        )
        self._startup_nodes_reachable = False
        self._initial_startup_nodes = list(self._startup_nodes)
        self._follow_cluster = follow_cluster
        self._skip_full_coverage_check = skip_full_coverage_check

    async def get_cluster_layout(
        self,
    ) -> tuple[list[ClusterNodeLocation], defaultdict[int, list[ClusterNodeLocation]]]:
        """
        Initializes the slots cache by asking all startup nodes what the
        current cluster configuration is.

        TODO: Currently the last node will have the last say about how the configuration is setup.
        Maybe it should stop to try after it have correctly covered all slots or when one node is
        reached and it could execute CLUSTER SLOTS command.
        """
        self._startup_nodes_reachable = False

        nodes_cache: dict[TCPLocation, ClusterNodeLocation] = {}
        tmp_slots: defaultdict[int, list[ClusterNodeLocation]] = defaultdict(list)

        all_slots_covered = False
        disagreements: list[str] = []
        replicas: set[str] = set()
        startup_node_errors: dict[Exception, list[str]] = {}

        nodes = self._initial_startup_nodes

        # With this option the client will attempt to connect to any of the previous set of nodes
        # instead of the original set of startup nodes
        if self._follow_cluster:
            nodes = self._startup_nodes

        for node in nodes:
            cluster_slots = {}
            if node:
                async with node.as_client(**self._connection_kwargs) as r:
                    try:
                        cluster_slots = await r.cluster_slots()
                        self._startup_nodes_reachable = True
                    except RedisError as err:
                        startup_node_errors.setdefault(err, []).append(node.name)
                        continue
            all_slots_covered = True

            for min_slot, max_slot in cluster_slots:
                _nodes = cluster_slots.get((min_slot, max_slot))
                if _nodes:
                    primary_node = ClusterNodeLocation(
                        host=_nodes[0]["host"],
                        port=_nodes[0]["port"],
                        server_type="primary",
                        node_id=_nodes[0]["node_id"],
                    )
                    replica_nodes = [
                        ClusterNodeLocation(
                            host=n["host"],
                            port=n["port"],
                            server_type="replica",
                            node_id=n["node_id"],
                        )
                        for n in _nodes[1:]
                    ]

                    primary_node.host = primary_node.host or node.host
                    nodes_cache[TCPLocation(primary_node.host, primary_node.port)] = primary_node

                    for i in range(min_slot, max_slot + 1):
                        if i not in tmp_slots:
                            tmp_slots[i].append(primary_node)
                            for replica_node in replica_nodes:
                                nodes_cache[TCPLocation(replica_node.host, replica_node.port)] = (
                                    replica_node
                                )
                                tmp_slots[i].append(replica_node)
                                replicas.add(replica_node.name)
                        else:
                            # Validate that 2 nodes want to use the same slot cache setup
                            if tmp_slots[i][0].name != node.name:
                                disagreements.append(
                                    f"{tmp_slots[i][0].name} vs {node.name} on slot: {i}",
                                )
                                if len(disagreements) > 5:
                                    raise RedisClusterError(
                                        "startup_nodes could not agree on a valid slots cache."
                                        f" {', '.join(disagreements)}"
                                    )
            if not (nodes_cache and tmp_slots):
                all_slots_covered = False
            elif not self._skip_full_coverage_check and (
                await self._cluster_require_full_coverage(nodes_cache)
            ):
                all_slots_covered = set(tmp_slots.keys()) == HASH_SLOTS_SET

            if all_slots_covered:
                break

        if not self._startup_nodes_reachable:
            startup_error = RedisClusterError(
                "Redis Cluster cannot be connected. Please provide at least one reachable node."
            )
            if startup_node_errors:
                startup_error.__cause__ = list(startup_node_errors.keys())[-1]
            raise startup_error
        if not all_slots_covered:
            raise RedisClusterError(
                "Not all slots are covered after attempting to query all startup nodes. "
                f"{len(tmp_slots)} of {HASH_SLOTS} covered."
            )

        self._startup_nodes.clear()
        self._startup_nodes.extend(nodes_cache.values())
        return list(nodes_cache.values()), tmp_slots

    async def _cluster_require_full_coverage(
        self, nodes: dict[TCPLocation, ClusterNodeLocation]
    ) -> bool:
        """
        If exists 'cluster-require-full-coverage no' config on redis servers,
        then even all slots are not covered, cluster still will be able to
        respond
        """

        for node in nodes.values():
            if await self._node_require_full_coverage(node):
                return True
        return False

    async def _node_require_full_coverage(self, node: ClusterNodeLocation) -> bool:
        async with node.as_client(**self._connection_kwargs) as r_node:
            try:
                with r_node.decoding(False):
                    node_config = await r_node.config_get(["cluster-require-full-coverage"])
                    return b"yes" in node_config.values()
            except ResponseError as err:
                warnings.warn(
                    "Unable to determine whether the cluster requires full coverage "
                    f"due to response error from `CONFIG GET`: {err}. To suppress this "
                    "warning use skip_full_coverage=True when initializing the client."
                )
                return False
            except ConnectionError:
                return False



from __future__ import annotations

from typing import cast

from coredis._utils import EncodingInsensitiveDict
from coredis.response._callbacks import ResponseCallback
from coredis.response._utils import flat_pairs_to_dict, flat_pairs_to_ordered_dict
from coredis.response.types import (
    StreamEntry,
    StreamInfo,
    StreamPending,
    StreamPendingExt,
)
from coredis.typing import (
    AnyStr,
    OrderedDict,
    ResponseType,
    StringT,
)


class StreamRangeCallback(
    ResponseCallback[list[list[StringT | list[StringT]]], tuple[StreamEntry, ...]]
):
    def transform(
        self,
        response: list[list[StringT | list[StringT]]],
    ) -> tuple[StreamEntry, ...]:
        return tuple(
            StreamEntry(cast(StringT, r[0]), flat_pairs_to_ordered_dict(cast(list[StringT], r[1])))
            for r in response
        )


class ClaimCallback(
    ResponseCallback[
        list[StringT | list[StringT | list[StringT]]], tuple[AnyStr, ...] | tuple[StreamEntry, ...]
    ]
):
    def transform(
        self,
        response: list[StringT | list[StringT | list[StringT]]],
    ) -> tuple[AnyStr, ...] | tuple[StreamEntry, ...]:
        if self.options.get("justid") is not None:
            return tuple(cast(list[AnyStr], response))
        else:
            return StreamRangeCallback()(cast(list[list[StringT | list[StringT]]], response))


class AutoClaimCallback(
    ResponseCallback[
        list[StringT | list[StringT | list[StringT | list[StringT]]]],
        tuple[AnyStr, tuple[AnyStr, ...]]
        | tuple[AnyStr, tuple[StreamEntry, ...], tuple[AnyStr, ...]],
    ]
):
    def transform(
        self, response: list[StringT | list[StringT | list[StringT | list[StringT]]]]
    ) -> (
        tuple[AnyStr, tuple[AnyStr, ...]]
        | tuple[AnyStr, tuple[StreamEntry, ...], tuple[AnyStr, ...]]
    ):
        if self.options.get("justid") is not None:
            return cast(AnyStr, response[0]), tuple(cast(list[AnyStr], response[1]))
        else:
            return (
                cast(AnyStr, response[0]),
                StreamRangeCallback()(cast(list[list[StringT | list[StringT]]], response[1])),
                tuple(cast(list[AnyStr], response[2])) if len(response) > 2 else (),
            )


class MultiStreamRangeCallback(
    ResponseCallback[
        dict[StringT, list[list[list[StringT] | StringT]]] | None,
        dict[AnyStr, tuple[StreamEntry, ...]] | None,
    ]
):
    def transform(
        self,
        response: dict[StringT, list[list[list[StringT] | StringT]]] | None,
    ) -> dict[AnyStr, tuple[StreamEntry, ...]] | None:
        if response:
            mapping: dict[AnyStr, tuple[StreamEntry, ...]] = {}

            for stream_id, entries in response.items():
                mapping[cast(AnyStr, stream_id)] = tuple(
                    StreamEntry(
                        cast(StringT, r[0]), flat_pairs_to_ordered_dict(cast(list[StringT], r[1]))
                    )
                    for r in entries
                )

            return mapping
        return None


class PendingCallback(
    ResponseCallback[
        list[StringT | int | list[StringT | int | list[StringT]] | None],
        StreamPending | tuple[StreamPendingExt, ...],
    ]
):
    def transform(
        self,
        response: list[StringT | int | list[StringT | int | list[StringT]] | None],
    ) -> StreamPending | tuple[StreamPendingExt, ...]:
        if not self.options.get("count"):
            consumers = [
                (cast(StringT, r[0]), int(r[1])) for r in cast(list[StringT], response[3] or [])
            ]
            return StreamPending(
                cast(int, response[0]),
                cast(StringT, response[1]),
                cast(StringT, response[2]),
                OrderedDict(consumers),
            )
        else:
            return tuple(
                StreamPendingExt(*(cast(tuple[StringT, StringT, int, int], tuple(sub[:4]))))
                for sub in cast(list[list[StringT | int]], response)
            )


class StreamInfoCallback(ResponseCallback[dict[StringT, ResponseType], StreamInfo]):
    def transform(
        self,
        response: dict[StringT, ResponseType],
    ) -> StreamInfo:
        res = EncodingInsensitiveDict(response)
        if not self.options.get("full"):
            k1 = "first-entry"
            kn = "last-entry"
            e1: StreamEntry | None = None
            en: StreamEntry | None = None

            if len(res.get(k1, []) or []) > 0:
                v = res.get(k1)
                e1 = StreamEntry(v[0], flat_pairs_to_ordered_dict(v[1]))
                res.pop(k1)

            if len(res.get(kn, []) or []) > 0:
                v = res.get(kn)
                en = StreamEntry(v[0], flat_pairs_to_ordered_dict(v[1]))
                res.pop(kn)
            res.update({"first-entry": e1, "last-entry": en})
        else:
            groups = res.get("groups")
            if groups:
                normalized_groups = []
                for group in groups:
                    g = EncodingInsensitiveDict(flat_pairs_to_dict(group))
                    consumers = g["consumers"]
                    normalized_consumers = []
                    for consumer in consumers:
                        normalized_consumers.append(flat_pairs_to_dict(consumer))
                    g["consumers"] = normalized_consumers
                    normalized_groups.append(g)
                res["groups"] = normalized_groups
            res.update(
                {
                    "entries": tuple(
                        StreamEntry(k[0], flat_pairs_to_ordered_dict(k[1]))
                        for k in res.get("entries", [])
                    )
                }
            )
        stream_info: StreamInfo = {
            "first-entry": res.get("first-entry"),
            "last-entry": res.get("last-entry"),
            "length": res["length"],
            "radix-tree-keys": res["radix-tree-keys"],
            "radix-tree-nodes": res["radix-tree-nodes"],
            "groups": res["groups"],
            "last-generated-id": res["last-generated-id"],
            "max-deleted-entry-id": str(res.get("max-deleted-entry-id")),
            "entries-added": int(res.get("entries-added", 0)),
            "recorded-first-entry-id": str(res.get("recorded-first-entry-id")),
            "entries-read": int(res.get("entries-read", 0)),
            "entries": res.get("entries"),
        }
        return stream_info
